#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains ArtellaDrive Python client implementation
"""

from __future__ import print_function, division, absolute_import

import os
import json
import time
import codecs
import socket
import base64
import random
import hashlib
import logging
import threading
import traceback
from collections import OrderedDict
try:
    from urllib.parse import urlparse, urlencode, urlunparse
    from urllib.request import urlopen, Request
    from urllib.error import HTTPError, URLError
except ImportError:
    from urlparse import urlparse, urlunparse
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError, URLError

from artella import dcc
from artella.core import consts, utils, exceptions, dccplugin

logger = logging.getLogger('artella')


class ArtellaDriveClient(object):
    """
    Class used to interact with ArtellaDrive application on client side
    """

    _challenge_path = None

    def __init__(self, host=consts.DEFAULT_HOST, port=consts.DEFAULT_PORT, extensions=None):
        self._host = host               # Contains default IP host used by the client.
        self._port = port               # Contains default port used by the client.
        self._auth_header = ''          # Contains authentication header read from challenge file.
        self._remote_sessions = list()  # Contains list of available remote sessions.
        self._local_projects = dict()   # Contains dictionary of all available user local Artella projects.
        self._remote_projects = dict()  # Contains a dictionary of all available user remote Artella projects.
        self._batch_ids = set()         # Contains a list that tracks all calls made to the client during a session.
        self._socket_buffer = None      # Contains instance of the socket buffer used to communicate with Artella App.
        self._running = False           # Flag that is True while Artella Drive thread is running.
        self._available = True          # Flag that is True while Artella Drive is available.
        self._exception = None          # Contains exception caused while Artella Drive is running.
        self._extensions = extensions   # Contains a list with all extensions managed by the client.
        self._client_thread = None      # Contains Python client thread

    # ==============================================================================================================
    # PROPERTIES
    # ==============================================================================================================

    @property
    def is_running(self):
        """
        Returns whether or not current Artella Drive thread is running.
        While Artella Drive thread is running, it listens responses from Artella Drive

        :return: True if the Artella Drive thread is running; False otherwise.
        :rtype: bool
        """

        return self._running

    @property
    def is_available(self):
        """
        Returns whether or not current Artella Drive client is available.
        Artella Drive Client is not available, for example, is the user closes Artella Drive App manually.

        :return: True if Artella Drive Client is available; False otherwise.
        :rtype: bool
        """

        return self._available

    # ==============================================================================================================
    # CLASS FUNCTIONS
    # ==============================================================================================================

    @classmethod
    def get(cls, extensions=None):
        """
        Returns a new instance of ArtellaDriveClient making sure that authorization is valid

        :param list(str) extensions: List of extensions
        :return: New instance of Artella Drive object
        :rtype: ArtellaDriveClient
        """

        artella_client = cls(extensions=extensions)

        # Challenge value gets updated when the Artella drive restarts.
        # We need to check this each time in case the local server got restarted
        auth_header = artella_client.update_auth_challenge()
        if not auth_header:
            logger.warning('Local Artella Drive is not available. Please launch Artella Drive App.')
            return None

        # Updates the list of available remotes
        artella_client.update_remotes_sessions(show_dialogs=False)

        return artella_client

    # ==============================================================================================================
    # CHALLENGE FILE
    # ==============================================================================================================

    def check(self, update=False):
        """
        Checks whether or not current client is available and running

        :param update: bool, Whether or not remote sessions should be updated
        :return: True if the client is available and running; False otherwise.
        :rtype: bool
        """

        return self._available and self._running and self.get_remote_sessions(update=update)

    def get_challenge_file_path(self):
        """
        Returns path where challenge file is located in local desktop machine

        :return: dictionary containing challenge file path
        :rtype: dict
        :example:
        >>> self.get_challenge_file_path()
        {
            "challenge_file_path": "C:\\users\\artella\\Application Data\\artella\\artella-challenge"
        }
        """

        req = Request('http://{}:{}/v2/localserve/auth/challenge-file-path'.format(self._host, self._port))
        rsp = self._communicate(req, skip_auth=True)

        return rsp

    def update_auth_challenge(self):
        """
        Collects some data from a local file to confirm we are running on the same that Artella drive is running.

        :return: string containing authentication header for current session or None if authentication was invalid
        :rtype: str or None
        """

        if not ArtellaDriveClient._challenge_path:
            rsp = self.get_challenge_file_path()
            if not rsp or 'error' in rsp:
                logger.error('Unable to get challenge file path "{}"'.format(rsp))
                self._auth_header = None
                return self._auth_header
            ArtellaDriveClient._challenge_path = rsp.get('challenge_file_path')

        if not os.path.exists(ArtellaDriveClient._challenge_path):
            logger.error('Challenge file not found: "{}"'.format(ArtellaDriveClient._challenge_path))
            self._auth_header = None
            return self._auth_header

        with open(os.path.expanduser(ArtellaDriveClient._challenge_path), 'rb') as fp:
            base_auth_header = fp.read()

        self._auth_header = consts.AUTH_HEADER.format(codecs.encode(base_auth_header[:64], 'hex').decode('ascii'))

        return self._auth_header

    def has_remote_sessions(self):
        """
        Returns whether or not current Artella Drive Client has remote sessions available or not
        :return: bool
        """

        return self._remote_sessions and len(self._remote_sessions) > 0

    def get_remote_sessions(self, update=False):
        """
        Returns a list with all available remote sessions

        :return: List of cached remote sessions in current Artella Drive Client
        :rtype: list(str)
        """

        if not self._remote_sessions and update:
            self.update_remotes_sessions()

        return self._remote_sessions

    def update_remotes_sessions(self, show_dialogs=True):
        """
        Collects all available remote servers available.

        :return: List of available remote sessions received from Artella Drive App.
        :rtype: list(str)
        """

        def _ping_remote_sessions():
            rsp = self.ping()
            if 'error' in rsp:
                msg = 'Attempting to retrieve remote sessions: "{}" '.format(rsp.get('error'))
                logger.error(msg)
            self._remote_sessions = rsp.get('remote_sessions', list())

        utils.clear_list(self._remote_sessions)

        if not self.update_auth_challenge():
            msg = 'Unable to authenticate to Artella Drive App.'
            logger.error(msg)
            if show_dialogs:
                dcc.show_error('Artella - Authentication error', msg)
            return

        _ping_remote_sessions()

        # If not remote sessions found, we recheck after 1 second to make sure Artella Drive App is running.
        if not self._remote_sessions:
            time.sleep(1)
            _ping_remote_sessions()
            if not self._remote_sessions:
                msg = 'No remote sessions available. Please visit your Project Drive in Artella Web App and try again!'
                logger.error(msg)

        return self._remote_sessions

    # ==============================================================================================================
    # SESSION
    # ==============================================================================================================

    def get_storage_id(self):
        """
        Returns storage ID of the machine this client is running on

        :return: ID indicating this desktop instance
        :rtype: str
        :example:
        >>> self.get_storage_id()
        "d7apalzl2rdnphe5wuccytqq3i"
        """

        req = Request('http://{}:{}/v2/localserve/kv/settings/machine-id'.format(self._host, self._port))
        storage_id = self._communicate(req)

        return storage_id

    def get_metadata(self):
        """
        Returns general data related with current session by asking remote server

        :return: Returns a dictionary containing all keys and values in kv namespace
        :rtype: dict
        :example:
        >>> self.get_metadata()
        {
            "machine-id": "d7apalzl2rdnphe5wuccytqq3i",
            "workspace": "C:/Users/artella/artella-files",
            "openers.log": "C:/Users/artella/AppData/Roaming/artella/openers.log"
        }
        """

        params = urlencode({'dump': 'true'})
        req = Request('http://{}:{}/v2/localserve/kv/settings?{}'.format(self._host, self._port, params))
        rsp = self._communicate(req)

        return rsp

    # ==============================================================================================================
    # PROJECTS
    # ==============================================================================================================

    def get_local_projects(self, force_update=False):
        """
        Returns all available project files found in the local user machine

        :param bool force_update: Whether to force the cache of the current available local projects
        :return: List of dictionaries containing all the information of the current available
        Artella on local user machine projects
        :rtype: dict(str, dict(str, str))
        """

        if self._local_projects and not force_update:
            return self._local_projects

        local_root = self.get_local_root()
        if not local_root or not os.path.isdir(local_root):
            logger.warning(
                'Impossible to retrieve current available projects because local root is not valid: {}'.format(
                    local_root))
            return list()

        project_files = utils.get_files(local_root, pattern=consts.PROJECT_METADATA_FILE)

        for project_file in project_files:
            project_dir = os.path.dirname(project_file)
            project_name = os.path.basename(project_dir)
            self._local_projects.setdefault(project_name, None)
            with open(project_file, 'r') as fh:
                lines = fh.readlines()
                project_data = {
                    'directory': project_dir
                }
                for line in lines:
                    line_split = line.split(': ')
                    if not line_split:
                        continue
                    project_data[line_split[0]] = line_split[1].strip()
                if project_data:
                    self._local_projects[project_name] = project_data

        return self._local_projects

    def get_remote_projects(self, force_update=False):
        """
        Returns all available projects in Artella

        :param bool force_update: Whether to force the cache of the current available remote projects
        :return: Dictionary that contains all projects for each one of the available remote sessions
        :rtype: dict(str, list(str))
        """

        if self._remote_projects and not force_update:
            return self._remote_projects

        if not self._remote_sessions:
            return dict()

        for remote_session in self._remote_sessions:
            remote_session_api = remote_session.get('api', None)
            if not remote_session_api:
                continue
            remote_session_projects = remote_session.get('projects', dict())
            self._remote_projects[remote_session_api] = remote_session_projects

        return self._remote_projects

    def get_project_name(self, project_id, remote_session=None):
        """
        Returns the project name giving its remote Artella ID

        :param str project_id: ID of the project in remote Artella server
        :param str remote_session: Remote session we want to search project on. If not given, the project will be
            search in all available remote sessions
        :return: Name of the project
        :rtype: str
        """

        remote_projects = self.get_remote_projects()
        if not remote_projects:
            logger.warning('No remote projects available to search project name for: "{}"'.format(project_id))
            return ''

        project_name = ''
        if remote_session:
            if remote_session not in remote_projects:
                logger.warning(
                    'Given remote session: "{}" is not available. Impossible to retrieve name for project: "{}"'.format(
                        remote_session, project_id))
                return project_name

            if project_id not in remote_projects[remote_session]:
                logger.warning('Project ID "{}" not found in remote session: "{}"'.format(
                    project_id, remote_session))
                return project_name

            project_name = remote_projects[remote_session][project_id].get('name', '')
        else:
            for remote_id, remote_data in remote_projects.items():
                for remote_project_id, project_data in remote_data.items():
                    if project_id != remote_project_id:
                        continue
                    project_name = project_data.get('name', '')
                    break

        if not project_name:
            logger.warning('No project name found for project: "{}"'.format(project_id))

        return project_name

    # ==============================================================================================================
    # PATHS
    # ==============================================================================================================

    def is_artella_path(self, file_path):
        """
        Returns whether or not given file path is an Artella file path or not
        A path is considered to be an Artella path if the path is located inside the Artella project folder
        in the user machine

        :param str file_path: path to check. If not given, current DCC scene file path will be used
        :return: True if the given file path is an Artella path; False otherwise.
        :rtype: bool
        """

        if not file_path or not os.path.exists(file_path):
            return False

        local_root = os.path.normpath(os.path.abspath(self.get_local_root()))
        file_path = os.path.normpath(os.path.abspath(file_path))

        return file_path.startswith(local_root)

    def get_local_root(self):
        """
        Returns the local storage root path for this machine by asking to remote server.
        .. note::
            If local root cannot be retrieved from server (this can happen because server is down or because user
            doesn't have a working internet connection) then this function will try to retrieve the local root path
            from ARTELLA_FOLDER_ROOT environment variable. This environment variable must be defined by users manually
            to point to local root path.

        :return: Absolute path where files are stored locally
        :rtype: str or None
        :example:
        >>> self.get_local_root()
        "C:/Users/artella/artella-files"
        """

        req = Request('http://{}:{}/v2/localserve/kv/settings/workspace'.format(self._host, self._port))
        local_root = self._communicate(req)
        if not local_root or (isinstance(local_root, dict) and 'error' in local_root):
            alr = os.environ.get(consts.ALR, None)
            if alr:
                logger.warning('Unable to get local storage root. Using env var instead: "{}"'.format(consts.ALR))
                local_root = alr
            else:
                logger.error('Unable to get local storage root.')
                logger.info(
                    'Check that the local Artella Drive service is running and you have a working internet connection.')
                logger.info(
                    'To work offline set the "{}" environment variable to the proper local project directory'.format(
                        consts.ALR))
        else:
            local_root = utils.clean_path(local_root)
            os.environ[consts.ALR] = local_root

        return local_root

    def resolve_path(self, path):
        """
        Converts a local path to a remote handle path representation and vice versa

        :param str path: local path or remote server path
        :return: Returns a dictionary containing resolved path or error message is the path cannot be resolve
        :rtype: dict
        :example:
        >>> os.environ['ART_LOCAL_ROOT'] = "C:/Users/artella/artella-files"
        >>> self.resolve_path("ART_LOCAL_ROOT/project/refs/ref.png")
        {
            'handle': 'project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png',
            'file_path': u'C:\\Users\\artella\\artella-files\\project\\refs\\ref.png'
        }
        >>> self.resolve_path("artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png")
        {
            'handle': 'project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png',
            'file_path': u'C:\\Users\\artella\\artella-files\\project\\refs\\ref.png'
        }
        >>> self.resolve_path("C:/Users/artella/artella-files/project/refs/ref.png")
        {
            'handle': 'project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png',
            'file_path': 'C:\\Users\\artella\\artella-files\\project\\refs\\ref.png'
        }
        >>> self.resolve_path("C:/Invalid/Path/ref.png")
        {
            'url': 'http://127.0.0.1:29282/v2/localserve/resolve',
            'error': 'Failed to reach the local ArtellaDrive: "Bad Request"'
        }
        """

        # We set request object depending whether given path is an URI one or not
        if is_uri_path(path):
            payload = {'handle': path}
        else:
            payload = {'file_path': path}

        req = Request('http://{}:{}/v2/localserve/resolve'.format(self._host, self._port))
        req.add_header('Content-Type', 'application/json')
        rsp = self._communicate(req, json.dumps(payload).encode())
        if 'error' in rsp:
            logger.error('Attempting to resolve "{}" "{}"'.format(path, rsp.get('error')))

        return rsp

    def translate_path(self, file_path):
        """
        Converts a file path to a local file path taking into account the available projects.

        :param str file_path: File path we want to to translate to its user local version
        :return: User local version of the given path
        :rtype: str
        :example:
        >>> self.translate_path("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        "C:/Users/Tomi/artella/data/ProjectA/Assets/Characters/A/Model/a.ma"
        >>> self.translate_path("/ProjectA/Assets/Characters/A/Model/a.ma")
        "C:/Users/Tomi/artella/data/ProjectA/Assets/Characters/A/Model/a.ma"
        >>> self.translate_path("$ART_LOCAL_ROOT/ProjectA/refs/ref.png")
        "C:/Users/Tomi/artella/data/ProjectA/refs/ref.png"
        """

        if is_uri_path(file_path):
            return self.resolve_path(file_path)

        path = file_path
        if not os.path.isabs(file_path):
            path = self.relative_path_to_absolute_path(file_path)
        path = utils.clean_path(os.path.expanduser(os.path.expandvars(path)))

        local_root = self.get_local_root()
        local_project_names = list((self.get_local_projects() or dict()).keys())

        for old_alr in consts.OLD_LOCAL_ROOTS:
            old_alr_str = '${}'.format(old_alr)
            if path.startswith(old_alr_str):
                path = utils.clean_path(path.replace(old_alr, local_root))
                path = path[1:] if path.startswith('$') else path

        path_split = path.split('/')
        total_chars = 0
        for split in path_split:
            if split in local_project_names:
                break
            total_chars += len(split) + 1

        common_path = path[total_chars:]
        translated_path = '{}/{}'.format(local_root, common_path)

        rsp = self.resolve_path(translated_path)
        if 'error' in rsp:
            logger.warning('Error while resolving translated file path: "{}"!'.format(path))
            return path

        if not translated_path.startswith(local_root):
            logger.warning('Was not possible to translate file path: "{}"!'.format(path))
            return path

        return translated_path

    def is_path_translated(self, file_path):
        """
        Returns whether or not given path is already translated to a valid Artella path

        :param file_path: str, path you want to check translation validation
        :return: True if the given path is an already translated Artella path; False otherwise
        :rtype: bool
        :example:
        >>> self.is_path_translated("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        False
        >>> self.is_path_translated("$ART_LOCAL_ROOT/ProjectA/refs/ref.png")
        True
        """

        if not file_path:
            return False

        return file_path.startswith('${}'.format(consts.ALR))

    def convert_path(self, file_path):
        """
        Converts given path to a path that Artella can understand

        :param str file_path: File path we want to convert
        :return: str
        :rtype: str
        :example:
        >>> self.translate_path("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        "$$ART_LOCAL_ROOT/ProjectA/Assets/Characters/A/Model/a.ma"
        >>> self.translate_path("/ProjectA/Assets/Characters/A/Model/a.ma")
        "$ART_LOCAL_ROOT/ProjectA/Assets/Characters/A/Model/a.ma"
        """

        local_root = self.get_local_root()

        path_to_convert = file_path
        if not os.path.isabs(file_path):
            path_to_convert = self.relative_path_to_absolute_path(file_path)
        translated_path = utils.clean_path(self.translate_path(path_to_convert))

        converted_path = translated_path
        if translated_path.startswith(local_root):
            converted_path = translated_path.replace(local_root, '${}'.format(consts.ALR))

        return converted_path
    
    def relative_path_to_absolute_path(self, relative_path, project_id=None):
        """
        Converts a relative path to an absolute local file path taking into account the given project id

        :param str relative_path: Relative path to a file
        :param str or None project_id: project id the relative path file belongs to
        :return: Absolute local file path of the relative path in the given project
        :rtype: str
        """

        if os.path.isabs(relative_path):
            return relative_path

        local_root = self.get_local_root()

        if not project_id:
            project_id = utils.split_path(relative_path)[0]

        local_project_names = self.get_local_projects() or dict()
        for project_name, project_data in local_project_names.items():
            if project_id == project_name:
                return utils.clean_path(os.path.join(local_root, relative_path))
            if project_id == project_data.get('remote', None):
                return utils.clean_path(os.path.join(local_root, project_name, relative_path))

        return relative_path

    def project_name_from_path(self, file_path):
        """
        Returns project name given file path belongs to. None if the path does not belongs to nay user local project.

        :param str file_path: path we want to retrieve project name of.
        :return: project name given path belongs to.
        :rtype: str
        """

        local_project_names = list((self.get_local_projects() or dict()).keys())

        project_name = None

        path_split = file_path.split('/')
        for split in path_split:
            if split in local_project_names:
                project_name = split
                break

        return project_name

    def project_id_from_path(self, file_path):
        """
        Returns project ID given file path belongs to. None if the path does not belongs to any user local project.

        :param str file_path: path we want to retrieve project ID of.
        :return: Project ID given path belongs to.
        :rtype: str
        """

        local_projects = self.get_local_projects() or dict()
        if not local_projects:
            logger.warning('No local projects available to retrieve project ID from path!')
            return None

        project_name = self.project_name_from_path(file_path)
        if not project_name:
            logger.warning('Was not possible to retrieve Project ID from path: {}!'.format(file_path))
            return None

        if project_name not in list(local_projects.keys()):
            logger.warning(
                'Was not possible to retrieve Project ID from path because project "{}" is not recognized!'.format(
                    project_name))
            return None

        project_data = local_projects[project_name]
        project_id = project_data.get('remote', None)

        return project_id

    # ==============================================================================================================
    # FILES/FOLDERS
    # ==============================================================================================================

    def exists_in_server(self, file_path):
        """
        Returns whether or not given file path exists in Artella remote server.

        :param str file_path: File path we want to check. It can be an absolute local file path or an URI path
        :return: True if the file exists in server; False otherwise.
        :rtype: bool
        """

        rsp = self.resolve_path(file_path)

        return False if 'error ' in rsp else True

    def status(self, file_paths, include_remote=False):
        """
        Returns the status of a file from the remote server

        :param str or list(str) file_paths: Local path(s) or Resolved URI path(s) of a folder/file or list of
            folders/files
        :param bool include_remote: Whether to include remote server information into the response. Take into account
            that this call can slow down the server response.
        :return:
        :rtype: dict
        :example:
        >>> self.status('artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png')
        >>> self.status('C:/Users/artella/artella-files/projects/refs/ref.png')
        {
            'project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png':
            {
                'local_info':
                {
                    'content_length': 390223,
                    'remote_version': 1,
                    'name': u'Area.png',
                    'modified_time': '2020-03-11T23:38:01.3985577+01:00',
                    'mode': u'0666',
                    'signature':
                    'sha1:f67817bdfa6a828ab3a80110d2e52ca8a430afb7',
                    'path': u'C:\\Users\\artella\\artella-files\\project\\refs\\ref.png'
                }
            }
        }
        """

        if not self.get_remote_sessions(update=True):
            logger.warning(
                'No remote sessions available. Artella App Drive status call aborted.')
            return list()

        file_paths = utils.force_list(file_paths, remove_duplicates=True)
        result = list()

        # Make sure that paths are clean
        file_paths = [utils.clean_path(file_path) for file_path in file_paths]

        for file_path in file_paths:
            try:
                file_path = file_path.decode('utf-8')
            except UnicodeDecodeError:
                file_path = file_path.decode('latin-1')
            uri_path = path_to_uri(file_path) if not is_uri_path(file_path) else file_path
            uri_parts = urlparse(uri_path)
            params = urlencode({'handle': uri_parts.path, 'include-remote': str(bool(include_remote)).lower()})
            req = Request('http://{}:{}/v2/localserve/fileinfo?{}'.format(
                self._host, self._port, params))
            rsp = self._communicate(req)
            if 'error' in rsp:
                logger.error(
                    'Attempting to retrieve status "{}" "{}"'.format(
                        uri_path.encode(), rsp.get('error').encode('utf-8')))
                continue

            result.append(rsp)

        return result

    def download(self, paths, version=None, recursive=False, overwrite=True, folders_only=False):
        """
        Downloads files from remote server

        :param list(str) paths: List of file/folder paths (full or URI paths) to download from remote server
        :param int version: If given, specified version of the file/folder will be downloaded
        :param bool recursive: If True and we download a folder, all folder contents will be downloaded also
        :param bool overwrite: If True, already downloaded files will be overwritten with the contents in remote server
        :param bool folders_only: If True, only folder structure will be downloaded
        :return: Dictionary containing batch ID of the operation
        :rtype: dict
        :example:
        >>> self.download('artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png')
        >>> self.download('C:/Users/artella/artella-files/project/refs/', recursive=True)
        >>> self.download('C:/Users/artella/artella-files/project/refs/', recursive=True, version=1)
        >>> self.download('C:/Users/artella/artella-files/project/refs/ref.png')
        {
            'batch_id': '1584199877-4'
        }
        >>> self.download('Non/Valid/Path')
        {}
        """

        path_handle_map = paths_to_handles(paths, as_dict=True)
        handles = list(path_handle_map.values())
        if not path_handle_map or not handles:
            return dict()
        logger.debug('Handles: "{}"'.format(handles))

        payload = {
            'handles': handles,
            'recursive': recursive,
            'replace_locals': overwrite,
            'folders_only': folders_only
        }
        if version is not None:
            payload['version'] = int(version)
        try:
            req = Request('http://{}:{}/v2/localserve/transfer/download'.format(self._host, self._port))
            rsp = self._communicate(req, json.dumps(payload).encode())
        except Exception:
            rsp = {'error': traceback.format_exc()}
        if 'error' in rsp:
            logger.warning(
                'Unable to download file paths "{}" "{}"'.format(list(path_handle_map.keys()), rsp.get('error')))
            return rsp

        return self._track_response(rsp)

    def pause_downloads(self):
        """
        Pauses current active downloads from Artella server

        :return: Response from Artella server after downloads are stopped
        :rtype: dict
        """

        req = Request('http://{}:{}/v2/localserve/pause/pause/downloads'.format(self._host, self._port))
        rsp = self._communicate(req)

        return rsp

    def pause_uploads(self):
        """
        Pauses current active uploads to Artella server

        :return: Response from Artella server after downloads are stopped
        :rtype: dict
        """

        req = Request('http://{}:{}/v2/localserve/pause/pause/uploads'.format(self._host, self._port))
        rsp = self._communicate(req)

        return rsp

    def resume_downloads(self):
        """
        Resume current paused downloads from Artella server

        :return: Response from Artella server after downloads are resumed
        :rtype: dict
        """

        req = Request('http://{}:{}/v2/localserve/pause/resume/downloads'.format(self._host, self._port))
        rsp = self._communicate(req)

        return rsp

    def resume_uploads(self):
        """
        Resume current paused uploads to Artella server

        :return: Response from Artella server after uploads are resumed
        :rtype: dict
        """

        req = Request('http://{}:{}/v2/localserve/pause/resume/uploads'.format(self._host, self._port))
        rsp = self._communicate(req)

        return rsp

    def upload(self, paths, folders_only=False, comment=''):
        """
        Uploads a new file/folder or a new version of an existing file/folder to the remote server

        :param list(str) paths: List of file/folder paths (full or URI paths) to upload to the remote server
        :param bool folders_only: If True, only folder structure will be uploaded
        :param str comment: Comment that will be linked to the uploaded files/folders
        :return: Response from Artella server after the upload operation is done.
        :rtype: dict
        >>> self.upload('artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png')
        >>> self.upload('C:/Users/artella/artella-files/project/refs/', folders_only=True, comment='new folder')
        >>> self.upload('C:/Users/artella/artella-files/project/refs/ref.png', comment='new file version')
        {
            'batch_id': '1584199877-4'
        }
        >>> self.upload('Non/Valid/Path')
        {}
        """

        paths = utils.force_list(paths, remove_duplicates=True)

        path_handle_map = paths_to_handles(paths, as_dict=True)
        handles = list(path_handle_map.values())
        if not path_handle_map or not handles:
            return dict()
        logger.debug('Handles: "{}"'.format(handles))

        payload = {
            'handles': handles,
            'recursive': False,
            'commit_message': comment or '',
            'folders_only': folders_only
        }
        req = Request('http://{}:{}/v2/localserve/transfer/upload'.format(self._host, self._port))

        rsp = self._communicate(req, json.dumps(payload).encode())
        if 'error' in rsp:
            logger.warning(
                'Unable to upload file paths "{}" "{}"'.format(list(path_handle_map.keys()), rsp.get('error')))
            return rsp

        return self._track_response(rsp)

    def lock_file(self, file_paths, note=consts.DEFAULT_LOCK_NOTE):
        """
        Lock given file path in the web platform so that other users know it is in use

        :param str or list(str) file_paths: Absolute local file path(s) to lock
        :param str or None note: Optional note to add to the lock operation
        :return: Returns True if the lock file was successfully locked or if the if the file was already locked
            by the same user; False otherwise
        :rtype: bool
        :example:
        >>> self.lock_file('C:/Users/artella/artella-files/project/refs/ref.png')
        True
        """

        file_paths = utils.force_list(file_paths, remove_duplicates=True)
        result = OrderedDict()

        for file_path in file_paths:
            payload = {
                'handle': file_path,
                'note': note or consts.DEFAULT_LOCK_NOTE
            }
            req = Request('http://{}:{}/v2/localserve/lock'.format(self._host, self._port))
            req.add_header('Content-Type', 'application/json')
            rsp = self._communicate(req, json.dumps(payload), method='POST')
            if isinstance(rsp, dict):
                if 'error' in rsp:
                    logger.error('Unable to lock file path "{}" "{}"'.format(file_path, rsp.get('error')))
                    return False
            else:
                logger.error('Unable to lock file path "{}" "{}"'.format(file_path, rsp))
                return False

            result[file_path] = rsp.get('response', False) and rsp.get('status_code', 0) == 200

        return result

    def unlock_file(self, file_paths):
        """
        Unlocks given file paths in the web platform so that other users can use it

        :param str or list(str) file_paths: Absolute local file path(s) to unlock
        :return: Returns True if the file was successfully unlocked; False otherwise
        :rtype: bool
        """

        file_paths = utils.force_list(file_paths, remove_duplicates=True)
        result = OrderedDict()

        for file_path in file_paths:
            payload = {
                'handle': file_path
            }
            req = Request('http://{}:{}/v2/localserve/lock'.format(self._host, self._port))
            req.add_header('Content-Type', 'application/json')
            rsp = self._communicate(req, json.dumps(payload).encode(), method='DELETE')
            if 'error' in rsp:
                logger.error('Unable to unlock file path "{}" "{}"'.format(file_path, rsp.get('error')))
                return result

            if isinstance(rsp, dict):
                result[file_path] = rsp.get('response', False) and rsp.get('status_code', 0) == 200
            else:
                # If we try to unlock files that are not remote yet the response will be an string telling us that
                # information. We consider that as a valid lock.
                result[file_path] = True

        return result

    def file_current_version(self, file_path, _status=None):
        """
        Returns current version of the given file

        :param str file_path: Absolute local file path to retrieve current local version of
        :param str _status: new file status
        :return: Current local version of the given file path
        :rtype: int
        """

        status = _status or self.status(file_path, include_remote=True)
        if not status:
            return None

        current_version = -1
        for file_status in status:
            for file_uri_path, file_status_data in file_status.items():
                if 'local_info' not in file_status_data or not file_status_data['local_info'] or \
                        'remote_info' not in file_status_data:
                    continue
                else:
                    current_version = file_status_data['remote_info'].get('version', 0)
                    break

        return current_version

    def file_latest_version(self, file_path, _status=None):
        """
        Returns latest version of the given file
        :param str file_path: Absolute local file path to retrieve latest server version of
        :return: int
        """

        status = _status or self.status(file_path, include_remote=True)

        latest_version = 0
        for file_status in status:
            for file_uri_path, file_status_data in file_status.items():
                if 'local_info' not in file_status_data or not file_status_data['local_info'] or \
                        'remote_info' not in file_status_data:
                    continue
                else:
                    latest_version = file_status_data['remote_info'].get('raw', dict()).get('highest_version', 0)
                    break

        return latest_version

    def file_is_latest_version(self, file_path):
        """
        Returns whether or not given local file path is updated to the latest version available in Artella server.

        :param str file_path: Absolute local file path or URI path we want to check version of
        :return: True if the file is updated to the latest version available; False otherwise.
        :rtype: bool
        """

        status = self.status(file_path, include_remote=True)

        local_version = self.file_current_version(file_path, _status=status)
        latest_version = self.file_latest_version(file_path, _status=status)

        return local_version == latest_version

    def can_lock_file(self, file_path):
        """
        Returns whether or not current opened DCC file (or given file path) can be locked or not
        A file only can be locked if it is not already locked by other user.

        :param str or None file_path: Absolute local file path we want to check if can be locked or not.
        :return: True if the file can be locked by current user; False otherwise.
        :rtype: bool
        """

        is_locked, is_locked_by_me, is_locked_by_name, remote_record_found = self.check_lock(file_path)
        if not remote_record_found:
            logger.info('File "{}" is not versioned yet. No need to lock'.format(file_path))
            return True

        if not is_locked or (is_locked and is_locked_by_me):
            return True

        return False

    def check_lock(self, file_path):
        """
        Returns whether or not the given file is locked and whether or not current user is the one that has the file
        locked.

        :param str file_path: Absolute local file path to check lock status of
        :return: Returns a tuple with the following fields:
            - is_locked: True if the given file is locked; False otherwise
            - is_locked_by_me: True if the given file is locked by current user; False otherwise
            - locked_by_name: Name of the user that currently has the file locked
            - remote_record_found: Indicates whether the request relates to an existing remote file record or not
        :rtype: tuple(bool, bool, str, bool)
        """

        payload = {
            'handle': file_path
        }
        req = Request('http://{}:{}/v2/localserve/lock'.format(self._host, self._port))
        req.add_header('Content-Type', 'application/json')
        rsp = self._communicate(req, json.dumps(payload).encode(), method='GET')
        if 'error' in rsp:
            logger.error('Unable to check lock status for file path "{}" "{}"'.format(file_path, rsp.get('error')))
            return False, False, '', False

        is_locked = rsp.get('is_locked', False)
        is_locked_by_me = rsp.get('is_locked_by_me', False)
        locked_by_name = rsp.get('locked_by_name', '')
        remote_record_found = rsp.get('remote_record_found', False)

        return is_locked, is_locked_by_me, locked_by_name, remote_record_found

    def get_progress(self):
        """
        Returns the download progress information of the current download operation from Artella server

        :return: A tuple containing the following information:
            - amount of done download operations
            - amount of total download operations in progress
            - amount of total download operations that are going to be done
            - amount of total bytes downloaded
            - amount of total bytes to download
        :rtype: int, int, int, int, int
        """

        tx_count_done = 0
        tx_count_total = 0
        tx_bytes_done = 0
        tx_bytes_total = 0
        batches_complete = set()

        batch_ids = list(self._batch_ids)[:]

        for batch_id in batch_ids:
            params = urlencode({'batch-id': batch_id, 'details': True})
            req = Request('http://{}:{}/v2/localserve/progress/summary?{}'.format(self._host, self._port, params))
            rsp = self._communicate(req)
            tcd = int(rsp.get('transfer_count_done', 0))
            tct = int(rsp.get('transfer_count_total', 0))
            tbd = int(rsp.get('transfer_bytes_download_done', 0))
            tbt = int(rsp.get('transfer_bytes_download_total', 0))
            if tcd == tct or tbd == tbt:
                batches_complete.add(batch_id)
                continue
            tx_count_done += tcd
            tx_count_total += tct
            tx_bytes_done += tbd
            tx_bytes_total += tbt

        self._batch_ids -= batches_complete

        progress_value = 1
        if tx_bytes_total > 0:
            progress_value = float(tx_bytes_done) / float(tx_bytes_total)
        elif tx_count_total > 0:
            progress_value = float(tx_count_done) / float(tx_count_total)

        return int(progress_value * 100), tx_count_done, tx_count_total, tx_bytes_done, tx_bytes_total

    # ==============================================================================================================
    # TEST
    # ==============================================================================================================

    def ping(self):
        """
        Test call that returns whether the Artella Drive is valid or not

        :return: Returns a success response or auth failure message
        :rtype: dict
        """

        req = Request('http://{}:{}/v2/localserve/ping'.format(self._host, self._port))
        rsp = self._communicate(req, skip_auth=True)

        return rsp

    # ==============================================================================================================
    # ARTELLA DRIVE APP
    # ==============================================================================================================

    def artella_drive_connect(self):
        """
        Connects to the local user Artella Drive App via web socket so we can listen for realtime events
        :return:
        """

        server_address = (self._host, self._port)
        logger.info('Connecting to Artella Drive App web socket: {}'.format(server_address))
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(server_address)
        except Exception as exc:
            logger.error('Artella Plugin failed to connect to the Artella Drive App web socket: {}'.format(exc))
            return

        self.artella_drive_send_request(sock)
        self._socket_buffer = SocketBuffer(sock)
        rsp = self.artella_drive_read_response()
        self._running = True

        return rsp

    def artella_drive_disconnect(self):
        logger.debug('Disconnecting Artella Drive Client ...')
        self._running = False

        self._disconnect()

        return True

    def artella_drive_listen(self):
        self.artella_drive_connect()
        if not self._socket_buffer:
            logger.error('Socket to Artella Drive not connected.')
            return
        self._client_thread = threading.Thread(target=self._pull_messages,).start()

    def artella_drive_send_request(self, sock):

        if not self.update_auth_challenge():
            logger.error('Unable to authenticate to Artella Drive App.')
            return

        extensions_str = ''
        for extension in self._extensions:
            if extension != '':
                extension = '{}&'.format(extension)
            if extension.startswith('*'):
                extension = extension[1:]
            if extension.startswith('.'):
                extension = '%2A{}'.format(extension)
            extensions_str += 'file-handler={}'.format(extension)
        if extensions_str:
            extensions_str = extensions_str[:-1]

        key, expected_key_response = make_ws_key()
        path = '/v2/localserve/ws' if not extensions_str else '/v2/localserve/ws?{}'.format(extensions_str)
        r = [
            'GET {} HTTP/1.1'.format(path),
            'Upgrade: websocket',
            'Connection: Upgrade',
            'Host: {}:{}'.format(self._host, self._port),
            'Sec-WebSocket-Key: {}'.format(key),
            'Sec-WebSocket-Version: 13',
            'Authorization: {}'.format(self._auth_header),
        ]
        rr = consts.CRLF.join(r) + consts.CRLF + consts.CRLF
        logger.debug(
            'Artella Plugin sending websocket access request to the local Artella Drive App: {}'.format(r))

        try:
            sock.sendall(rr.encode())
        except Exception as exc:
            logger.warning('Was not possible to send request to Artella Drive App: {}'.format(exc))
            return False

        return True

    def artella_drive_read_response(self):
        logger.debug('Reading socket response ...')
        line = ''
        while line != consts.CRLF:
            line = self._socket_buffer.read_line()

        return line

    # ==============================================================================================================
    # INTERNAL
    # ==============================================================================================================

    def _communicate(self, req, data=None, skip_auth=False, method=None):
        """
        Internal function that sends a request to ArtellaDriver server

        :param Request req: URL request object
        :param str data: additional encoded data to send to the server
        :param bool skip_auth: whether authorization check should be skip or not
        :param str or None method: Method that should be used bu urrllib to send the request
        :return: dict, dictionary containing the answer from the server.
            If the call is not valid, a dictionary containing the url and error message is returned.
        """

        logger.debug('Making request to Artella Drive "{}"'.format(req.get_full_url()))
        if data:
            logger.debug('Request payload dump: "{}" | {}'.format(json.loads(data), type(data)))

        if not self._auth_header and not skip_auth:
            rsp = self.update_auth_challenge()
            if not rsp:
                msg = 'Unable to authenticate'
                logger.error(msg)
                return {'error': msg, 'url': req.get_full_url()}

        req.add_header('Authorization', self._auth_header)
        if method is not None:
            req.get_method = lambda: method
        try:
            rsp = urlopen(req, data)
        except URLError as exc:
            if hasattr(exc, 'reason'):
                msg = 'Failed to reach the local ArtellaDrive: "{}"'.format(exc.reason)
            elif hasattr(exc, 'code'):
                msg = 'ArtellaDrive is unable to fulfill the request "{}"'.format(exc.code)
            else:
                msg = exc
            logger.debug(exc)
            logger.error(msg)
            return {'error': msg, 'url': req.get_full_url()}
        else:
            raw_data = rsp.read()
            if not raw_data:
                return {'error': 'No Artella data response.', 'url': req.get_full_url()}
            else:
                try:
                    raw_data = "".join(chr(x) for x in bytearray(raw_data))
                    json_data = json.loads(raw_data)
                except ValueError:
                    logger.debug('ArtellaDrive data response: "{}"'.format(raw_data))
                    return raw_data
                except Exception as exc:
                    logger.error(exc)
                    return raw_data
                else:
                    logger.debug('ArtellaDriver JSON response: "{}"'.format(json_data))
                    return json_data

    def _disconnect(self):
        """
        Internal function that closes current Artella Drive client web socket

        :return: True if the web socket was successfully disconnected; False otherwise.
        :rtype: bool
        """

        try:
            if self._socket_buffer:
                self._socket_buffer.close()
                self._socket_buffer = None
                self._running = False
        except Exception as exc:
            logger.error(
                'Error while disconnect Artella Plugin from Artella Drive App web socket ... {}'.format(exc))
            return False

        return True

    def _track_response(self, rsp):
        """
        Internal function that extracts batch_id and transfer_data from response so that we can continue the
        tracking process

        :param dict rsp: Dictionary returned by a call to remote server
        :return: String containing ID of the operation
        :rtype: str
        """

        batch_id = rsp.get('batch_id')
        self._batch_ids.add(batch_id)

        return rsp

    def _pull_messages(self):
        logger.debug('Listening for commands on websocket')
        while self._running:
            try:
                msg = self._get_message()
            except Exception as exc:
                self._available = False
                try:
                    if exc.errno == 10054:
                        self._disconnect()
                except Exception as exc:
                    try:
                        logger.debug(str(exc))
                    except Exception as exc:
                        pass
                continue

            self._available = True
            logger.debug('Message received: {}'.format(msg))
            dccplugin.DccPlugin().pass_message(msg)

        # If Artella Drive Client is nor running we force disconnection
        # self._disconnect()

    def _get_message(self):
        op_code = ord(self._socket_buffer.get_char())
        v = ord(self._socket_buffer.get_char())
        if op_code != 129:
            raise exceptions.ArtellaException('Not a final text frame: {}'.format(op_code))
        if v < 126:
            length = v
        elif v == 126:
            a = ord(self._socket_buffer.get_char()) << 8
            b = ord(self._socket_buffer.get_char())
            length = a + b
        elif v == 127:
            # 8 byte payload length - we do not have any of these
            raise exceptions.ArtellaException('Unsupported payload length')
        else:
            raise exceptions.ArtellaException('Bad payload length: {}'.format(v))

        payload = self._socket_buffer.get_chars(length)

        return json.loads(payload)


class SocketBuffer(object):
    def __init__(self, sock):
        super(SocketBuffer, self).__init__()

        self._sock = sock
        self._buffer = ''

    def read_line(self):
        line = ''
        while True:
            c = self.get_char()
            line += c
            if c == '\n':
                return line

    def get_char(self):
        if len(self._buffer) == 0:
            self._buffer = self._sock.recv(consts.DEFAULT_BUFFER_SIZE)
        r = self._buffer[0]
        self._buffer = self._buffer[1:]

        return chr(r) if type(r) == int else r

    def get_chars(self, count):
        cc = ''
        for x in range(count):
            cc += self.get_char()

        return cc

    def close(self):
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()


def make_ws_key():
    """
     https://tools.ietf.org/html/rfc6455
    :return:
    """

    key = base64.b64encode(str(int(random.random() * 9999999)).encode()).decode()
    h = hashlib.sha1()
    h.update((key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").encode())
    expected_key_response = base64.b64encode(h.digest()).decode()

    return key, expected_key_response


def is_uri_path(file_path):
    """
    Returns whether or not given file path is using Artella URI schema or not

    :param str file_path: File path to check.
    :return: True if given path is using Artella URI schema; False otherwise.
    :rtype: bool
    :example:
    >>> is_uri_path("artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png")
    True
    >>> is_uri_path("C:/Users/artella/artella-files/project/refs/ref.png")
    False
    """

    if not file_path:
        return False

    uri = urlparse(file_path)

    return uri.scheme == consts.ARTELLA_URI_SCHEME


def path_to_uri(path):
    """
    Converts a path to a path that uses Artella API scheme

    :param str path: File to convert to Artella URI scheme format
    :return: Returns a path converted into Artella URI scheme format. If the conversion cannot be done, the path
        without any change will be returned.
    :rtype: str
    :example:
    >>> path_to_uri("artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png")
    "artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png"
    >>> path_to_uri("C:/Users/artella/artella-files/project/refs/ref.png")
    "artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png"
    >>> path_to_uri("Invalid/Path")
    "Invalid/Path"
    """

    path_uri = paths_to_uri(path)

    return path_uri[0]


def paths_to_uri(paths):
    """
    Converts a list of paths to paths that use Artella URI scheme

    :param str or list(str) paths: List of paths to convert to Artella URI scheme
    :return: List of paths converted into Artella URI scheme. If a path cannot be converted, it will be returned
        without any changes.
    :rtype: list(str)
    """

    paths = utils.force_list(paths)

    fixed_paths = list()

    for pth in paths:
        full_path = os.path.abspath(os.path.expandvars(os.path.expanduser(pth)))
        rsp = ArtellaDriveClient.get().resolve_path(full_path)
        if 'error' in rsp:
            logger.warning('Unable to translate path "{}" "{}"'.format(pth, rsp.get('error')))
            fixed_paths.append(pth)
            continue

        url_parts = (consts.ARTELLA_URI_SCHEME, '', rsp.get('handle'), '', '', '')
        fixed_path = urlunparse(url_parts)
        if not is_uri_path(fixed_path):
            logger.error('Failed to translate "{}" to URI: "{}"'.format(pth, fixed_path))
            fixed_paths.append(pth)
            continue

        fixed_paths.append(fixed_path)

    return fixed_paths


def path_to_handle(path):
    """
    Converts a path to an Artella handle path

    :param path: str, Path to convert to handle path, it can be an URI path or an absolute local file path
    :return: Returns an Artella handle path
    :rtype: str
    :example:
    >>> path_to_handle("artella:project__fo0pohsa2sea4wyr5zmzcwnzse/refs/ref.png")
    "file__12345"
    >>> path_to_handle("C:/Users/artella/artella-files/project/refs/ref.png")
    "file__12345"
    """

    handles = paths_to_handles(path, as_dict=False)

    return handles[0] if handles else path


def paths_to_handles(paths, as_dict=False):
    """
    Converts a list of paths (full paths or URI paths) to Artella handle paths

    :param str or list(str) paths:
    :param bool as_dict: If True, a dictionary mapping file paths with their handle will be returned.
    :return: List of handles associated to given file paths
    :rtype: list(str) or dict(str, str)
    """

    paths = utils.force_list(paths)

    handles = set()
    path_handle_map = dict()

    for pth in paths:
        if not is_uri_path(pth):
            uri_path = path_to_uri(pth)
            # If converted URI path is the same as the original path means that path is not valid, so we skip it
            if uri_path == pth:
                continue
        else:
            uri_path = pth
        uri_parts = urlparse(uri_path)
        handle = uri_parts.path
        handles.add(handle)
        path_handle_map[pth] = handle

    if as_dict:
        return path_handle_map

    return list(handles)


if __name__ == '__main__':

    from artella import loader
    loader.init()

    artella_cli = ArtellaDriveClient.get()
    print(artella_cli.ping())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains Artella DCC Plugin functionality
"""

from __future__ import print_function, division, absolute_import

import os
import time
import logging
import threading

from artella import dcc
from artella.core.dcc import parser
from artella.core import utils, qtutils, callbacks, splash
from artella.widgets import snackbar

if qtutils.QT_AVAILABLE:
    from artella.externals.Qt import QtCore, QtWidgets

logger = logging.getLogger('artella')

_DCC_PLUGIN = None


class _MetaDccPlugin(type):

    def __call__(cls, *args, **kwargs):

        global _DCC_PLUGIN
        if _DCC_PLUGIN:
            return _DCC_PLUGIN

        if dcc.is_maya():
            from artella.dccs.maya import plugin
            _DCC_PLUGIN = type.__call__(plugin.ArtellaMayaPlugin, *args, **kwargs)
        elif dcc.is_unreal():
            from artella.dccs.unreal import plugin
            _DCC_PLUGIN = type.__call__(plugin.ArtellaUnrealPlugin, *args, **kwargs)
        else:
            _DCC_PLUGIN = type.__call__(BaseArtellaDccPlugin, *args, **kwargs)

        return _DCC_PLUGIN


class BaseArtellaDccPlugin(object):

    MENU_NAME = 'Artella'
    _ASYNC_INVOKER, _SYNC_INVOKER = range(2)

    def __init__(self, artella_drive_client):
        super(BaseArtellaDccPlugin, self).__init__()

        self._artella_drive_client = artella_drive_client
        self._dev = False
        self._main_menu = None

        self._main_thread_invoker, self._main_thread_async_invoker = self._create_main_thread_invokers()

    # ==============================================================================================================
    # PROPERTIES
    # ==============================================================================================================

    @property
    def dev(self):
        """
        Returns whether Artella DCC plugin is being executed in dev mode or not
        :return: True if dev mode is enabled; False otherwise.
        :rtype: bool
        """

        return self._dev

    # ==============================================================================================================
    # INITIALIZATION / SHUTDOWN
    # ==============================================================================================================

    def setup_project(self, artella_local_root_path):
        """
        Setup Artella local root as current DCC active project
        This function should be override in specific DCC plugin implementation
        Is not an abstract function because its implementation is not mandatory

        :param str artella_local_root_path: current user Artella local root path
        """

        pass

    def get_version_variable_name(self):
        """
        Returns the environment variable name used to store current Arqtella DCC plugin version

        :return: Environment version name
        :rtype: str
        """

        return 'ARTELLA_{}_PLUGIN'.format(dcc.name())

    def get_version(self, force_update=False):
        """
        Returns current DCC plugin version

        :param bool force_update: Where or not force the update of the current Artella DCC plugin version
        :return: Version in string format (MAJOR.MINOR.PATH) of the current Artella DCC plugin
        :rtype: str or None
        """

        version_var = self.get_version_variable_name()
        plugin_version = os.environ.get(version_var, None)

        return plugin_version

    def init(self, dev=False, show_dialogs=True, create_menu=True, create_callbacks=True,  *args, **kwargs):
        """
        Initializes Artella plugin in current DCC.

        :param bool dev: Whether plugin is initialized in development mode or not
        :param bool show_dialogs: Whether dialogs should appear during plugin initialization or not
        :param bool create_menu: Whether menu should be created or not
        :param bool create_callbacks: Whether or not DCC callbacks should be created
        :return: True if the initialization was successful; False otherwise.
        :rtype: bool
        """

        self._dev = dev

        # Initialize DCC plugin version
        self.get_version(force_update=True)

        # Initialize Artella callbacks
        if create_callbacks:
            self.setup_callbacks()

        # Create Artella DCC Menu
        if create_menu:
            dcc.execute_deferred(self.create_menus)

        # Initialize Artella Drive client
        _init_client = kwargs.pop('init_client', True)
        if _init_client:
            self.init_client(show_dialogs=show_dialogs)

        logger.debug('trying to create quit signal ...')
        if qtutils.QT_AVAILABLE:
            app = QtWidgets.QApplication.instance()
            logger.debug('connecting dcc app to signal: {}'.format(app))
            if app:
                app.aboutToQuit.connect(self._on_quit_app)
            else:
                logger.debug('Impossible to connect because app does not exists')
        else:
            logger.debug('Impossible to create signal because qt is not available')

        return True

    def init_client(self, show_dialogs=True):
        """
        Initializes Artella Drive Client.

        :param bool show_dialogs: Whether dialogs should appear during plugin initialization or not
        :return: True if the Artella Drive client initialization was successful; False otherwise.
        :rtype: bool
        """

        artella_drive_client = self._artella_drive_client or self.get_client(show_dialogs=show_dialogs)
        if artella_drive_client:
            self.setup_project(artella_drive_client.get_local_root())
            artella_drive_client.artella_drive_listen()
        else:
            logger.warning(
                'Artella Drive Client was not initialized. Artella server '
                'dependant functionality will not be available!')
            return False

        return True

    def shutdown(self, dev=False):
        """
        Shutdown/Uninitialize Artella plugin in current DCC
        :return: True if the shutdown was successful; False otherwise
        :rtype: bool
        """

        # Remove Artella callbacks
        self.remove_callbacks()

        # Remove Artella DCC Menu
        if dev:
            self.remove_menus()
        else:
            dcc.execute_deferred(self.remove_menus)

        if not self._artella_drive_client:
            return False

        self._artella_drive_client.artella_drive_disconnect()
        self._artella_drive_client = None

        return True

    # ==============================================================================================================
    #  EXECUTION
    # ==============================================================================================================

    def execute_in_main_thread(self, fn, *args, **kwargs):
        """
        Executes the given function in the main thread when called from a non-main thread. This will block the
        calling thread until the function returns.

        .. warning: This can introduce a deadlock if the main thread is waiting for a background thread and the
            this thread is invoking this method. Since the main thread is waiting for the background thread to finish,
            Qt's event loop won't be able to process the request to execute in the main thread
        :param fn: function to call
        :param args:
        :param kwargs:
        :return:
        """

        return self._execute_in_main_thread(self._SYNC_INVOKER, fn, *args, **kwargs)

    def async_execute_in_main_thread(self, fn, *args, **kwargs):
        """
        Executes the given function in the main thread when called from a non-main thread. This call will return
        immediately and will not wait for the code to be executed in the main thread.

        :param fn: function to call
        :param args:
        :param kwargs:
        :return:
        """

        return self._execute_in_main_thread(self._ASYNC_INVOKER, fn, *args, **kwargs)

    # ==============================================================================================================
    #  MENU
    # ==============================================================================================================

    def get_main_menu(self):
        """
        Returns instance of Artella main menu
        :return:
        """

        return self._main_menu

    def create_menus(self):
        """
        Setup DCC Artella menu.
        If the menu already exists, it will be deleted and recreated.
        :return: True if the menu was created successfully; False otherwise
        :rtype: bool
        """

        if dcc.check_menu_exists(self.MENU_NAME):
            dcc.remove_menu(self.MENU_NAME)

        artella_menu = dcc.add_menu(self.MENU_NAME, icon='artella.png')
        if not artella_menu:
            return False

        self._main_menu = artella_menu

        return True

    def remove_menus(self):
        """
        Removes already created DCC Artella menu
        :return: True if the menu was removed successfully; False otherwise
        :rtype: bool
        """

        if not dcc.check_menu_exists(self.MENU_NAME):
            return False

        dcc.remove_menu(self.MENU_NAME)

        return True

    # ==============================================================================================================
    # CALLBACKS
    # ==============================================================================================================

    def setup_callbacks(self):
        """
        Setup DCC Artella callbacks
        :return:
        """

        callbacks.initialize_callbacks()

    def remove_callbacks(self):
        """
        Removes all DCC Artella callbacks previously created
        :return:
        """

        callbacks.uninitialize_callbacks()

    def validate_environment_for_callback(self, callback_name):
        """
        Checks that all necessary parts are available before executing a Maya callback

        :param str callback_name: name of the callback to validate
        """

        logger.info('validate_environment_for_callback for {}'.format(callback_name))
        client = self.get_client()
        local_root = dcc.clean_path()

    # ==============================================================================================================
    # ARTELLA DRIVE APP
    # ==============================================================================================================

    def update_auth_challenge(self):
        """
        Updates the authentication header by checking challenge file (if available)

        :return: True if the authenticator header was read successfully; False otherwise.
        :rtype: bool
        """

        if not self._artella_drive_client:
            return False

        auth_header = self._artella_drive_client.update_auth_challenge()

        return True if auth_header else False

    def get_client(self, show_dialogs=True):
        """
        Returns current Artella Drive Client being used by Artella Plugin

        :return: Instance of current ArtellaDriveClient being used; None if not Artella Drive Client is being used.
        :rtype: ArtellaDriveClient or None
        """

        # To avoid cyclic imports
        from artella.core import client

        if not self._artella_drive_client:
            # TODO: Here we are not taking into account custom extensions (check loader). We should store them
            # TODO: in an env variable and access them here
            dcc_extensions = dcc.extensions()
            artella_drive_client = client.ArtellaDriveClient.get(extensions=dcc_extensions)
            if not artella_drive_client:
                if show_dialogs:
                    self.show_warning_message(
                        'Local Drive Client not available. Please launch Artella Drive App', 'Artella Drive App')
                return None
            else:
                self._artella_drive_client = artella_drive_client
                self.init_client()
        else:
            if not self._artella_drive_client.is_running:
                self.init_client()

        if not self._artella_drive_client.is_available:
            self._artella_drive_client.update_remotes_sessions(show_dialogs=False)
            if not self._artella_drive_client.is_available:
                if show_dialogs:
                    self.show_warning_message(
                        'Local Drive Client not available. Please launch Artella Drive App', 'Artella Drive App')
                return None

        # The challenge value gets updated when the Artella Drive App restarts.
        # We must check auth each time in case Artella Drive is restarted but Maya didn't
        self.update_auth_challenge()

        return self._artella_drive_client

    def pass_message(self, json_data):
        """
        Executes handle message Artella plugin functionality in specific DCC thread to make sure that it does not
        conflicts with main DCC UI thread
        :param json_data:

        """

        logger.debug('Passing message to {}: {}'.format(dcc.name(), json_data))
        self.execute_in_main_thread(self.handle_message, json_data)

    def handle_message(self, msg):
        """
        Internal function that handles the response received from Artella Drive App

        :param dict msg: Dictionary containing the response from Artella server
        """

        logger.debug('Handling realtime message: {}'.format(msg))
        if not isinstance(msg, dict):
            logger.warning('Malformed realtime message: {}'.format(msg))
            return

        command_name = msg.get('type')
        dcc_name = dcc.name()

        if command_name.startswith(dcc_name):
            dcc_operation = command_name.split('-')[-1]
            file_path = msg['data']['ARTELLA_FILE']
            if dcc_operation == 'import':
                self._handle_import_message(file_path)
            elif dcc_operation == 'reference':
                self._handle_reference_message(file_path)
        elif command_name == 'authorization-ok':
            logger.debug('websocket connection successful.')
        elif command_name == 'open':
            file_path = msg['data']['ARTELLA_FILE']
            self._handle_open_message(file_path)
        else:
            pass

    def artella_info(self):
        """
        Function that prints useful information related with current Artella Plugin status
        """

        logger.info('Artella Plugin Status Info ...')
        artella_drive_client = self.get_client()
        logger.info(artella_drive_client.get_metadata())
        logger.info(artella_drive_client.get_storage_id())
        logger.info(artella_drive_client.ping())
        logger.info(artella_drive_client.artella_drive_connect())
        logger.info('Local Root: {}'.format(artella_drive_client.get_local_root()))

    # ==============================================================================================================
    # FILE PATHS
    # ==============================================================================================================

    def is_artella_path(self, file_path=None):
        """
        Returns whether or not given file path is an Artella file path or not
        A path is considered to be an Artella path if the path is located inside the Artella project folder
        in the user machine

        :param str file_path: path to check. If not given, current DCC scene file path will be used
        :return: True if the given file path is an Artella path; False otherwise.
        :rtype: bool
        """

        artella_drive_client = self.get_client(show_dialogs=False)
        if not artella_drive_client:
            return False

        file_path = file_path or dcc.scene_name()

        return artella_drive_client.is_artella_path(file_path)

    def local_path_to_uri(self, file_path):
        """
        Translates a local file path to its URI format
        :param str file_path: Absolute local file path we want to translate to URI
        :param str prefix:

        :return: path in its URI format if current DCC supports this feature; path without any change otherwise
        :rtype: str
        """

        if not dcc.supports_uri_scheme():
            logger.warning('Current DCC {} does not supports Artella URI scheme!'.format(dcc.name()))
            return file_path

        return file_path

    def translate_path(self, file_path):
        """
        Converts a file path to a local file path taking into account the available projects.

        :param str file_path: File path we want to to translate to its user local version
        :return: User local version of the given path
        :rtype: str
        :example:
        >>> self.translate_path("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        "C:/Users/Tomi/artella/data/ProjectA/Assets/Characters/A/Model/a.ma"
        >>> self.translate_path("/ProjectA/Assets/Characters/A/Model/a.ma")
        "C:/Users/Tomi/artella/data/ProjectA/Assets/Characters/A/Model/a.ma"
        """

        artella_drive_client = self.get_client(show_dialogs=False)
        if not artella_drive_client:
            return file_path

        return artella_drive_client.translate_path(file_path)

    def is_path_translated(self, file_path):
        """
        Returns whether or not given path is already translated to a valid Artella path

        :param file_path: str, path you want to check translation validation
        :return: True if the given path is an already translated Artella path; False otherwise
        :rtype: bool
        :example:
        >>> self.is_path_translated("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        False
        >>> self.is_path_translated("$ART_LOCAL_ROOT/ProjectA/refs/ref.png")
        True
        """

        artella_drive_client = self.get_client(show_dialogs=False)
        if not artella_drive_client:
            return False

        return artella_drive_client.is_path_translated(file_path)

    def convert_path(self, file_path):
        """
        Converts given path to a path that Artella can understand

        :param str file_path: File path we want to convert
        :return: str
        :rtype: str
        :example:
        >>> self.translate_path("C:/Users/Bobby/artella-files/ProjectA/Assets/Characters/A/Model/a.ma")
        "$$ART_LOCAL_ROOT/ProjectA/Assets/Characters/A/Model/a.ma"
        >>> self.translate_path("/ProjectA/Assets/Characters/A/Model/a.ma")
        "$ART_LOCAL_ROOT/ProjectA/Assets/Characters/A/Model/a.ma"
        """

        artella_drive_client = self.get_client(show_dialogs=False)
        if not artella_drive_client:
            return file_path

        return artella_drive_client.convert_path(file_path)

    @utils.timestamp
    def update_paths(self, file_path=None, show_dialogs=True, call_post_function=True, skip_save=True):
        """
        Updates all file paths of the given file path to make sure that they point to valid Artella file paths

        :param str file_path:
        :param bool show_dialogs:
        :param bool call_post_function:
        :param bool skip_save:
        :return:
        """

        artella_drive_client = self.get_client(show_dialogs=False)
        if not artella_drive_client:
            return False

        if not file_path:
            file_path = dcc.scene_name()
        if not file_path:
            msg = 'No file paths given to convert. Impossible to update paths.'
            logger.warning(msg)
            if show_dialogs:
                dcc.show_warning(title='Artella - Failed to update paths', message=msg)
            return False

        file_paths = utils.force_list(file_path, remove_duplicates=True)

        converted_paths = list()
        for file_path in file_paths:

            local_path = artella_drive_client.translate_path(file_path)

            ext = os.path.splitext(local_path)[-1]
            if ext not in dcc.extensions() or not os.path.isfile(local_path):
                logger.info('Skipping non DCC scene file path from conversion: "{}"'.format(local_path))
                continue

            can_lock = self.can_lock_file(local_path, show_dialogs=False)
            if not can_lock:
                logger.warning(
                    'File "{}" cannot be locked and paths cannot be updated. Skipping ...'.format(local_path))
                continue

            if dcc.scene_name() != file_path:
                dcc.open_scene(file_path, save=True)

            dcc_parser = parser.Parser()
            valid_convert, updated_paths = dcc_parser.update_paths(local_path)
            updated_paths = utils.force_list(updated_paths)
            converted_paths.extend(updated_paths)

            if valid_convert and updated_paths and not skip_save:
                dcc.save_scene()

        converted_paths = list(set(converted_paths))
        if call_post_function and converted_paths:
            self._post_update_paths(files_updated=converted_paths)

    # ==============================================================================================================
    # VERSIONS
    # ==============================================================================================================

    def make_new_version(self, file_path=None, comment=None, do_lock=False):
        """
        Uploads a new file/folder or a new version of current opened DCC scene file

        :param str file_path: Optional path of the file we want to create new version of. If not given, current
            opened DCC scene file path will be used.
        :param str comment: Optional comment to add to new version metadata. If not given, a generic message will be
            used.
        :param bool do_lock: Whether or not to force the lock of the file to make a new version. With new Artella
            version this is not mandatory.
        :return: True if the make new version operation is completed successfully; False otherwise.
        :rtype: bool
        """

        artella_drive_client = self.get_client()
        if not artella_drive_client:
            return False

        if not file_path:
            file_path = dcc.scene_name()
            if not file_path:
                msg = 'Please open a file before creating a new version'
                logger.warning(msg)
                return False

        can_lock = artella_drive_client.can_lock_file(file_path=file_path)
        if not can_lock:
            msg = 'Unable to lock file to make new version. File is already locked by other user.'
            dcc.show_error('File already locked by other user', msg)
            logger.error(msg)
            return False

        version_created = True

        comment = str(comment) if comment else 'New file version'

        file_version = artella_drive_client.file_current_version(file_path)
        if file_version is None:
            self.show_warning_message('Unable to retrieve version from current scene')
            return False

        next_version = file_version + 1

        is_locked, _, _, _ = artella_drive_client.check_lock(file_path)
        if not is_locked and do_lock:
            valid_lock = self.lock_file()
            if not valid_lock:
                self.show_error_message('Unable to lock file to make new version ({})'.format(next_version))
                return False

        logger.info('Saving current scene: {}'.format(file_path))
        valid_save = dcc.save_scene()
        if not valid_save:
            self.show_error_message('Unable to save current scene: "{}"'.format(file_path))
            version_created = False
        else:
            uri_path = self.local_path_to_uri(file_path)
            rsp = artella_drive_client.upload(uri_path, comment=comment)
            if rsp.get('error'):
                msg = 'Unable to upload a new version of file: "{}"\n{}\n{}'.format(
                    os.path.basename(file_path), rsp.get('url'), rsp.get('error'))
                self.show_error_message(msg)
                version_created = False

        if not is_locked and do_lock:
            self.unlock_file(show_dialogs=False)

        return version_created

    # ==============================================================================================================
    # LOCK/UNLOCK STATUS
    # ==============================================================================================================

    def check_lock(self, file_path=None, show_dialogs=True):
        """
        Returns whether or not the given file is locked and whether or not current user is the one that has the file
        locked.

        :param str file_path: Absolute local file path to check lock status of
        :return: Returns a tuple with the following fields:
            - is_locked: True if the given file is locked; False otherwise
            - is_locked_by_me: True if the given file is locked by current user; False otherwise
            - locked_by_name: Name of the user that currently has the file locked
            - remote_record_found: Indicates whether the request relates to an existing remote file record or not
        :param bool show_dialogs: Whether UI dialogs should appear or not.
        :rtype: tuple(bool, bool, str)
        """

        artella_drive_client = self.get_client(show_dialogs=show_dialogs)
        if not artella_drive_client:
            return False, False, '', False

        if not file_path:
            file_path = dcc.scene_name()
        if not file_path:
            msg = 'File "{}" does not exists. Impossible to check lock status!'.format(file_path)
            logger.warning(msg)
            if show_dialogs:
                dcc.show_warning(title='Artella - Failed to check lost status', message=msg)
            return False, False, '', False

        return artella_drive_client.check_lock(file_path)

    def can_lock_file(self, file_path=None, show_dialogs=True):
        """
        Returns whether or not current opened DCC file can locked or not
        A file only can be locked if it is not already locked by other user.

        :param str or None file_path: Absolute local file path we want to check if can be locked or not. If not given,
            current DCC scene file will be locked.
        :param bool show_dialogs: Whether UI dialogs should appear or not.
        :return: True if the file can be locked by current user; False otherwise.
        :rtype: bool
        """

        artella_drive_client = self.get_client(show_dialogs=show_dialogs)
        if not artella_drive_client:
            return False

        if not file_path:
            file_path = dcc.scene_name()
        if not file_path:
            msg = 'File "{}" does not exists. Impossible to check if file can be locked or not!'.format(file_path)
            logger.warning(msg)
            if show_dialogs:
                dcc.show_warning(title='Artella - Failed to check lost status', message=msg)
            return False

        return artella_drive_client.can_lock_file(file_path)

    def lock_file(self, file_path=None, force=False, show_dialogs=True):
        """
        Locks given file path in Artella Drive.

        :param str or None file_path: Absolute local file path we want to lock. If not given, current DCC scene file
            will be locked.
        :param bool force: Whether to force the lock operation. If the file is locked by other user, the lock is break.
        :param bool show_dialogs: Whether UI dialogs should appear or not.
        :return: True if the lock operation was successful; False otherwise
        :rtype: bool
        """

        artella_drive_client = self.get_client(show_dialogs=show_dialogs)
        if not artella_drive_client:
            return False

        if not file_path:
            file_path = dcc.scene_name()
            if not file_path:
                msg = 'Unable to get file name, has it been created!?'
                logger.warning(msg)
                if show_dialogs:
                    dcc.show_warning(title='Artella - Failed to lock file', message=msg)
                return False

        if not file_path or not os.path.isfile(file_path):
            msg = 'File "{}" does not exists. Impossible to lock.'.format(file_path)
            logger.warning(msg)
            if show_dialogs:
                dcc.show_error('Artella - Failed to lock File', msg)
            return False

        file_version = artella_drive_client.file_current_version(file_path)
        if file_version <= 0:
            logger.info('File "{}" is not versioned yet. No need to lock.'.format(file_path))
            return True

        is_locked, is_locked_by_me, is_locked_by_name, remote_record_found = artella_drive_client.check_lock(file_path)
        can_write = os.access(file_path, os.W_OK)
        if not can_write and is_locked_by_me:
            logger.warning('Unable to determine local write permissions for file: "{}"'.format(file_path))
        if is_locked and not is_locked_by_me:
            msg = 'This file is locked by another user ({}). The file must be unlocked in order to save a new version.'
            logger.warning(msg)
            if show_dialogs:
                dcc.show_warning('Artella - Failed to lock file', msg)
            return False
        elif force or not is_locked:
            msg = '"{}" needs to be locked in order to save your file. ' \
                  'Would you like to lock the file now?'.format(os.path.basename(file_path))
            result = True
            if show_dialogs:
                result = dcc.show_question('Artella - lock file', msg, cancel=False)
            if result is not True:
                return False

        valid_lock = artella_drive_client.lock_file(file_path)
        if not valid_lock:
            msg = 'Failed to lock "{}"'.format(file_path)
            logger.warning(msg)
            dcc.show_warning('Artella - Failed to lock file', msg)
            return False

        return True

    def unlock_file(self, file_path=None, show_dialogs=True, force=False):
        """
        Unlocks given file path in Artella Drive.

        :param str or None file_path: Absolute local file path we want to lock. If not given, current DCC scene file
            will be unlocked.
        :param bool show_dialogs: Whether UI dialogs should appear or not.
        :param bool force: Whether or not unlock operation should be done without asking the user
        :return: True if the lock operation was successful; False otherwise
        :rtype: bool
        """

        artella_drive_client = self.get_client(show_dialogs=show_dialogs)
        if not artella_drive_client:
            return False

        if not file_path:
            file_path = dcc.scene_name()
            if not file_path:
                msg = 'Unable to get file name, has it been created!?'
                logger.warning(msg)
                if show_dialogs:
                    dcc.show_warning(title='Artella - Failed to lock file', message=msg)
                return False

        if not file_path or not os.path.isfile(file_path):
            msg = 'File "{}" does not exists. Impossible to unlock.'.format(file_path)
            logger.warning(msg)
            if show_dialogs:
                dcc.show_error('Artella - Failed to unlock file', msg)
            return False

        result = True
        if show_dialogs and not force:
            msg = 'You have file "{}" locked in Artella.\nUnlock it now?'.format(os.path.basename(file_path))
            result = dcc.show_question('Artella - Unlock File', msg, cancel=False)
        if result is not True:
            return False

        uri_path = self.local_path_to_uri(file_path)
        valid_unlock = artella_drive_client.unlock_file(uri_path)
        if not valid_unlock:
            msg = 'Failed to unlock the file: "{}"\nTry unlocking it from the Artella ' \
                  'Drive area in the web browser'.format(os.path.basename(file_path))
            logger.warning(msg)
            if show_dialogs:
                dcc.show_error('Artella - Failed to unlock file', msg)
            return False

        return True

    # ==============================================================================================================
    # DOWNLOAD
    # ==============================================================================================================

    def download_file(self, file_path, show_dialogs=True):
        """
        Downloads a file from Artella server

        :param str file_path: File we want to download from Artella server
        :param bool show_dialogs: Whether UI dialogs should appear or not.
        :return: True if the download operation was successful; False otherwise.
        :example:
        >>> self.download_file("C:/Users/artella/artella-files/ProjectA/Assets/test/model/hello.a")
        True
        """

        artella_drive_client = self.get_client(show_dialogs=show_dialogs)
        if not artella_drive_client:
            return False

        local_path = artella_drive_client.translate_path(file_path)

        remote_file_size = 0
        file_status = artella_drive_client.status(local_path, include_remote=True)
        if not file_status:
            logger.warning('File "{}" not found in remote'.format(file_status))
            return False
        file_status = file_status[0]
        for version_id, version_data in file_status.items():
            if not version_id or not version_id.startswith('project__'):
                continue
            remote_file_size = version_data.get('remote_info', dict()).get('content_length', 0)
            break
        if not remote_file_size:
            logger.warning('File "{}" in remote has not a valid file size: {}'.format(local_path, remote_file_size))
            return False

        if show_dialogs:
            dcc_progress_bar = splash.ProgressSplashDialog()
            dcc_progress_bar.start()
            if qtutils.QT_AVAILABLE:
                QtWidgets.QApplication.instance().processEvents()

        artella_drive_client.download(local_path)
        time.sleep(1.0)

        valid_download = True
        while True:
            if show_dialogs:
                dcc_progress_bar.repaint()
                if dcc_progress_bar.is_cancelled():
                    artella_drive_client.pause_downloads()
                    valid_download = False
                    break
            progress, fd, ft, bd, bt = artella_drive_client.get_progress()
            progress_status = '{} | {} of {} KiB downloaded\n{} of {} files downloaded'.format(
                    os.path.basename(file_path), int(bd / 1024), int(bt / 1024), fd, ft)
            if show_dialogs:
                dcc_progress_bar.set_progress_value(value=progress, status=progress_status)
                if qtutils.QT_AVAILABLE:
                    QtWidgets.QApplication.instance().processEvents()
            if progress >= 100 or bd == bt:
                break

        if show_dialogs:
            dcc_progress_bar.end()

        total_checks = 0
        if valid_download:
            while not os.path.exists(local_path) and total_checks < 5:
                time.sleep(1.0)
                total_checks += 1
            if not os.path.exists(local_path):
                return False
            total_checks = 0
            file_size = os.path.getsize(local_path)
            while file_size != remote_file_size and total_checks < 5:
                file_size = os.path.getsize(local_path)
                time.sleep(1.0)
                total_checks += 1
            if file_size != remote_file_size:
                return False

        return valid_download

    # ==============================================================================================================
    # UI
    # ==============================================================================================================

    def show_artella_message(self, text, title='', duration=None, closable=True):
        """
        Shows an Artella message

        :param text: str, artella text to show
        :param title: str, title of the artella message
        :param duration: float or None, if given, message only will appear the specified seconds
        :param closable: bool, Whether the message can be closed by the user or not (when duration is given)
        """

        logger.debug(str(text))
        if not qtutils.QT_AVAILABLE:
            return

        snackbar.SnackBarMessage.artella(text=text, title=title, duration=duration, closable=closable)

    def show_success_message(self, text, title='', duration=None, closable=True):
        """
        Shows a success message

        :param text: str, success text to show
        :param title: str, title of the success message
        :param duration: float or None, if given, message only will appear the specified seconds
        :param closable: bool, Whether the message can be closed by the user or not (when duration is given)
        """

        logger.info(str(text))
        if not qtutils.QT_AVAILABLE:
            return

        snackbar.SnackBarMessage.success(text=text, title=title, duration=duration, closable=closable)

    def show_info_message(self, text, title='', duration=None, closable=True):
        """
        Shows an info message

        :param text: str, info text to show
        :param title: str, title of the info message
        :param duration: float or None, if given, message only will appear the specified seconds
        :param closable: bool, Whether the message can be closed by the user or not (when duration is given)
        """

        logger.info(str(text))
        if not qtutils.QT_AVAILABLE:
            return

        snackbar.SnackBarMessage.info(text=text, title=title, duration=duration, closable=closable)

    def show_warning_message(self, text, title='', duration=None, closable=True):
        """
        Shows a warning message

        :param text: str, warning text to show
        :param title: str, title of the warning message
        :param duration: float or None, if given, message only will appear the specified seconds
        :param closable: bool, Whether the message can be closed by the user or not (when duration is given)
        """

        logger.warning(str(text))
        if not qtutils.QT_AVAILABLE:
            return

        snackbar.SnackBarMessage.warning(text=text, title=title, duration=duration, closable=closable)

    def show_error_message(self, text, title='', duration=None, closable=True):
        """
        Shows an error message

        :param text: str, error text to show
        :param title: str, title of the error message
        :param duration: float or None, if given, message only will appear the specified seconds
        :param closable: bool, Whether the message can be closed by the user or not (when duration is given)
        :return:
        """

        logger.error(str(text))
        if not qtutils.QT_AVAILABLE:
            return

        snackbar.SnackBarMessage.error(text=text, title=title, duration=duration, closable=closable)

    # ==============================================================================================================
    # INTERNAL
    # ==============================================================================================================

    def _post_update_paths(self, **kwargs):
        """
        Internal function that is called after update paths functionality is over. Can be override in custom DCC
        plugins.
        """

        pass

    def _handle_open_message(self, file_path):
        """
        Internal function that is called when websocket receives a message to open a file in DCC

        :param str file_path: Absolute local file path we need to open in current DCC
        :return:
        """

        return dcc.open_scene(file_path=file_path, save=False)

    def _handle_import_message(self, file_path):
        """
        Internal function that is called when websocket receives a message to import a file in DCC

        :param str file_path: Absolute local file path we need to import into current DCC
        :return:
        """

        return dcc.import_scene(file_path=file_path)

    def _handle_reference_message(self, file_path):
        """
        Internal function that is called when websocket receives a message to reference a file in DCC

        :param str file_path: Absolute local file path we need to reference into current DCC
        :return:
        """

        return dcc.reference_scene(file_path=file_path)

    def _execute_in_main_thread(self, invoker_id, fn, *args, **kwargs):
        """
        Internal function that executes the given function and arguments with the given invoker. If the invoker
        is not ready or if the calling thread is the main thread, the function is called immediately with given
        arguments.

        :param int invoker_id: _SYNC_INVOKER or _ASYNC_INVOKER
        :param callable fn: function to call
        :param args:
        :param kwargs:
        :return: Return value from the invoker
        :rtype: object
        """

        dcc_main_thread_fn = dcc.pass_message_to_main_thread_fn()
        if not qtutils.QT_AVAILABLE and not dcc_main_thread_fn:
            return fn(*args, **kwargs)

        # If DCC has a specific function to invoke functions in main thread, we use it
        if dcc_main_thread_fn:
            return dcc_main_thread_fn(fn, *args, **kwargs)
        else:
            invoker = (
                self._main_thread_invoker if invoker_id == self._SYNC_INVOKER else self._main_thread_async_invoker)
            if invoker:
                if QtWidgets.QApplication.instance() and (
                        QtCore.QThread.currentThread() != QtWidgets.QApplication.instance().thread()):
                    return invoker.invoke(fn, *args, **kwargs)
                else:
                    return fn(*args, **kwargs)
            else:
                return fn(*args, **kwargs)

    def _create_main_thread_invokers(self):
        """
        Internal function that creates invoker objects that allow to invoke function calls on the main thread when
        called from a different thread
        """

        invoker = None
        async_invoker = None

        if qtutils.QT_AVAILABLE:

            class MainThreadInvoker(QtCore.QObject):
                """
                Class that implements a mechanism to execute a function with arbitrary arguments in main thread for DCCs
                that support Qt
                """

                def __init__(self):
                    super(MainThreadInvoker, self).__init__()

                    self._lock = threading.Lock()
                    self._fn = None
                    self._res = None

                def invoke(self, fn, *args, **kwargs):
                    """
                    Invoke the given function with the given arguments and keyword arguments in the main thread

                    :param function fn: function to execute in main thread
                    :param tuple args: args for the function
                    :param dict kwargs: Named arguments for the function
                    :return: Returns the result returned by the function
                    :rtype: object
                    """

                    # Acquire lock to make sure that both function and result are not overwritten by synchronous calls
                    # to this method from different threads
                    self._lock.acquire()

                    try:
                        self._fn = lambda: fn(*args, **kwargs)
                        self._res = None

                        # Invoke the internal function that will run the function.
                        # NOTE: We cannot pass/return arguments through invokeMethod as
                        # this isn't properly supported by PySide
                        QtCore.QMetaObject.invokeMethod(self, '_do_invoke', QtCore.Qt.BlockingQueuedConnection)

                        return self._res
                    finally:
                        self._lock.release()

                @QtCore.Slot()
                def _do_invoke(self):
                    """
                    Internal function that executes the function
                    """

                    self._res = self._fn()

            class MainThreadAsyncInvoker(QtCore.QObject):
                """
                Class that implements a mechanism to execute a function with arbitrary arguments in main
                thread asynchronously
                for DCCs that support Qt
                """

                __signal = QtCore.Signal(object)

                def __init__(self):
                    super(MainThreadAsyncInvoker, self).__init__()

                    self.__signal.connect(self.__execute_in_main_thread)

                def invoke(self, fn, *args, **kwargs):
                    """
                     Invoke the given function with the given arguments and keyword arguments in the main thread

                     :param function fn: function to execute in main thread
                     :param tuple args: args for the function
                     :param dict kwargs: Named arguments for the function
                     :return: Returns the result returned by the function
                     :rtype: object
                     """

                    self._signal.emit(lambda: fn(*args, **kwargs))

                def __execute_in_main_thread(self, fn):
                    """
                    Internal function that executes the function
                    """

                    fn()

            # Make sure invoker exists in main thread
            invoker = MainThreadInvoker()
            async_invoker = MainThreadAsyncInvoker()
            if QtCore.QCoreApplication.instance():
                invoker.moveToThread(QtCore.QCoreApplication.instance().thread())
                async_invoker.moveToThread(QtCore.QCoreApplication.instance().thread())

        return invoker, async_invoker

    # ==============================================================================================================
    # CALLBACKS
    # ==============================================================================================================

    def _on_quit_app(self):
        """
        Internal callback function is called when a DCC with Qt support is closed. We use it to make sure that
        Artella Drive Client is stopped.
        """

        logger.debug('Quiting app ...')

        artella_drive_client = self.get_client()
        if not artella_drive_client:
            return

        artella_drive_client.artella_drive_disconnect()

    # ==============================================================================================================
    # ABSTRACT
    # ==============================================================================================================

    @utils.abstract
    def register_uri_resolver(self):
        """
        Function that registers DCC specific Artella URI resolver
        This function must be implemented in those dcc.that support this feature
        """

        pass


@utils.add_metaclass(_MetaDccPlugin)
class DccPlugin(object):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains DCC reroute functions
"""

from __future__ import print_function, division, absolute_import

from artella.core.dcc import reroute
from artella.core.utils import abstract

main = __import__('__main__')


class Dccs(object):
    Standalone = 'standalone'
    Maya = 'maya'
    Max = 'max'
    Houdini = 'houdini'

    packages = {
        'cmds': Maya,
        'pymxs': Max,
        'hou': Houdini,
    }


def is_standalone():
    """
    Check if current environment is standalone or not
    :return: bool
    """

    return not any(pkg in list(main.__dict__.keys()) for pkg in Dccs.packages)


def is_maya():
    """
    Checks if Maya is available or not
    :return: bool
    """

    return 'cmds' in main.__dict__


def is_max():
    """
    Checks if Max is available or not
    :return: bool
    """

    return 'MaxPlus' in main.__dict__ or 'pymxs' in main.__dict__


def is_mobu():
    """
    Checks if MotionBuilder is available or not
    :return: bool
    """

    return 'pyfbsdk' in main.__dict__


def is_houdini():
    """
    Checks if Houdini is available or not
    :return: bool
    """

    return 'hou' in main.__dict__


def is_unreal():
    """
    Checks if Unreal is available or not
    :return: bool
    """

    try:
        import unreal
    except ImportError:
        return False

    return True


@reroute
@abstract
def name():
    """
    Returns name of current DCC

    :return: Returns name of DCC without any info about version
    :rtype: str
    """

    pass


@reroute
@abstract
def nice_name():
    """
    Returns nice name of current DCC

    :return: Returns formatted DCC name
    :rtype: str
    """

    pass


@reroute
@abstract
def version():
    """
    Returns version of DCC application

    :return: Returns integer number indicating the version of the DCC application
    :rtype: int
    """

    pass


@reroute
@abstract
def extensions():
    """
    Returns a list of available extension for DCC application

    :return: List of available extensions with the following format: .{EXTENSION}
    :rtype: list(str)
    """

    pass


@reroute
@abstract
def scene_name():
    """
    Returns the name of the current scene

    :return: Full file path name of the current scene. If no file is opened, None is returned.
    :rtype: str or None
    """

    pass


@reroute
@abstract
def new_scene(force=True, **kwargs):
    """
    Creates a new scene inside DCC
    :param force: True to skip saving of the current opened DCC scene; False otherwise.
    :return: True if the new scene is created successfully; False otherwise.
    :rtype: bool
    """

    pass


@reroute
@abstract
def scene_is_modified():
    """
    Returns whether or not current opened DCC file has been modified by the user or not

    :return: True if current DCC file has been modified by the user; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def open_scene(file_path, save=True):
    """
    Opens DCC scene file
    :param str file_path: Absolute local file path we want to open in current DCC
    :param bool save: Whether or not save current opened DCC scene file
    :return: True if the save operation was successful; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def save_scene(force=True, **kwargs):
    """
    Saves DCC scene file

    :param bool force: Whether to force the saving operation or not
    :return: True if the save operation was successful; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def import_scene(file_path):
    """
    Opens scene file into current opened DCC scene file
    :param str file_path: Absolute local file path we want to import into current DCC
    :return: True if the import operation was successful; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def reference_scene(file_path, **kwargs):
    """
    References scene file into current opened DCC scene file
    :param str file_path: Absolute local file path we want to reference into current DCC
    :return: True if the reference operation was successful; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def supports_uri_scheme():
    """
    Returns whether or not current DCC support URI scheme implementation

    :return: True if current DCC supports URI scheme implementation; False otherwise
    """

    pass


@reroute
@abstract
def pass_message_to_main_thread_fn():
    """
    Returns function used by DCC to execute a function in DCC main thread in the next idle event of that thread.

    :return If DCC API supports it, returns function to call a function in main thread from other thread
    """

    pass


@reroute
@abstract
def is_batch():
    """
    Returns whether or not current DCC is being executed in batch mode (no UI)
    :return: True if current DCC session is being executed in batch mode; False otherwise.
    :rtype: bool
    """

    pass


@reroute
@abstract
def clean_path(file_path):
    """
    Cleans given path so it can be properly used by current DCC

    :param str file_path: file path we want to clean
    :return: Cleaned version of the given file path
    :rtype: str
    """

    pass


@reroute
@abstract
def get_installation_paths(versions=None):
    """
    Returns installation path of the given versions of current DCC

    :param list(int) versions: list of versions to find installation paths of. If not given, current DCC version
        installation path will be returned
    :return: List of installation paths of the given DCC versions
    :rtype: list(str)
    """

    pass


@reroute
@abstract
def is_udim_path(file_path):
    """
    Returns whether or not given file path is an UDIM one

    :param str file_path: File path we want to check
    :return: True if the given paths is an UDIM path; False otherwise.
    :rtype: bool
    """

    pass


@reroute
@abstract
def execute_deferred(fn):
    """
    Executes given function in deferred mode (once DCC UI has been loaded)
    :param callable fn: Function to execute in deferred mode
    :return:
    """

    pass


@reroute
@abstract
def register_dcc_resource_path(resources_path):
    """
    Registers path into given DCC so it can find specific resources
    :param resources_path: str, path we want DCC to register
    """

    pass


@reroute
@abstract
def main_menu_toolbar():
    """
    Returns Main menu toolbar where DCC menus are created by default

    :return: Native object that represents main menu toolbar in current DCC
    :rtype: object
    """

    pass


@reroute
@abstract
def get_menus():
    """
    Return all the available menus in current DCC. This function returns specific DCC objects that represents DCC
    UI menus.

    :return: List of all menus names in current DCC
    :rtype: list(str)
    """

    pass


@reroute
@abstract
def get_menu_items():
    """
    Returns all available menu items in current DCC. This function returns specific DCC objects that represents DCC
    UI menu items.

    :return: List of all menu item names in current DCC
    :rtype: list(str)
    """

    pass


@reroute
@abstract
def get_menu(menu_name):
    """
    Returns native DCC menu with given name
    :param str menu_name: name of the menu to search for
    :return: Native DCC menu object or None if the menu does not exists
    :rtype: str or None
    """

    pass


@reroute
@abstract
def get_menu_item(menu_item_name):
    """
    Returns native DCC menu item with given name
    :param str menu_item_name: name of the menu item to search for
    :return: Native DCC menu object or None if the menu does not exists
    :rtype: str or None
    """

    pass


@reroute
@abstract
def check_menu_exists(menu_name):
    """
    Returns whether or not menu with given name exists

    :param str menu_name: name of the menu to search for
    :return: True if the menu already exists; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def add_menu(menu_name, parent_menu=None, tear_off=True, icon='', **kwargs):
    """
    Creates a new DCC menu.

    :param str menu_name: name of the menu to create
    :param object parent_menu: parent menu to attach this menu into. If not given, menu will be added to
    specific DCC main menu toolbar. Must be specific menu DCC native object
    :param bool tear_off: Whether or not new created menu can be tear off from its parent or not
    :param str icon: name of the icon to be used in this menu
    :return: New DCC native menu object created or None if the menu item was not created successfully
    :rtype: object or None
    """

    pass


@reroute
@abstract
def remove_menu(menu_name):
    """
    Removes menu from current DCC if exists

    :param str menu_name: name of the menu to remove
    :return: True if the removal was successful; False otherwise
    :rtype: bool
    """

    pass


@reroute
@abstract
def add_menu_item(menu_item_name, menu_item_command='', parent_menu=None, icon='', **kwargs):
    """
    Adds a new menu item to the given parent menu. When the item is clicked by the user the given command will be+
    executed
    :param str menu_item_name: name of the menu item to create
    :param str menu_item_command: command to execute when menu item is clicked
    :param object parent_menu: parent menu to attach this menu into. Must be specific menu DCC native object
    :param str icon: name of the icon to be used in this menu item
    :return: New DCC native menu item object created or None if the menu item was not created successfully
    :rtype: object or None
    """

    pass


@reroute
@abstract
def add_sub_menu_item(menu_item_name, menu_item_command='', parent_menu=None, icon='', **kwargs):
    """
    Adds a new sub menu item to the given parent menu. When the item is clicked by the user the given command will be+
    executed
    :param str menu_item_name: name of the menu item to create
    :param str menu_item_command: command to execute when menu item is clicked
    :param object parent_menu: parent menu to attach this menu into. Must be specific menu DCC native object
    :param str icon: name of the icon to be used in this menu item
    :return: New DCC native menu item object created or None if the menu item was not created successfully
    :rtype: object or None
    """

    pass


@reroute
@abstract
def remove_menu_item(menu_item_name, parent_menu):
    """
    Removes a menu item from the given parent menu.
    :param str menu_item_name: name of the menu item to remove
    :param object parent_menu: parent menu to remove this menu from. Must be specific menu DCC native object
    :return: Trie if the operation was successful; False otherwise.
    :rtype: bool
    """

    pass


@reroute
@abstract
def add_menu_separator(parent_menu):
    """
    Adds a new separator to the given parent menu
    :param object parent_menu: parent menu to attach this menu into. Must be specific menu DCC native object
    """

    pass


@reroute
@abstract
def get_main_window():
    """
    Returns Qt object that references to the main DCC window we are working on

    :return: An instance of the current DCC Qt main window
    :rtype: QMainWindow or QWidget or None
    """

    pass


@reroute
@abstract
def show_info(title, message):
    """
    Shows an information dialog that users need to accept/reject.

    :param str title: text that is displayed in the title bar of the dialog
    :param str message: text which is shown to the user telling them what operation they need to confirm
    :return: True if the user accepted the operation; False otherwise.
    :rtype: bool
    """

    pass


@reroute
@abstract
def show_question(title, message, cancel=True):
    """
    Shows a question message box that can be used to show question text to users.

    :param str title: text that is displayed in the title bar of the dialog
    :param str message: text which is shown to the user telling them what operation they need to confirm
    :param bool cancel: Whether or not cancel button should appear in question message box
    :return: True if the user presses the Ok button; False if the user presses the No button; None if the user preses
        the Cancel button
    :rtype: bool or None
    """

    pass


@reroute
@abstract
def show_warning(title, message, print_message=False):
    """
    Shows a warning message box that can be used to show warning text to users.

    :param str title: text that is displayed in the title bar of the dialog
    :param str message: default text which is placed n the plain text edit
    :param bool print_message: whether or not print message in DCC output command window
    """

    pass


@reroute
@abstract
def show_error(title, message, print_message=False):
    """
    Shows an error message box that can be used to show critical text to users.

    :param str title: text that is displayed in the title bar of the dialog
    :param str message: default text which is placed n the plain text edit
    :param bool print_message: whether or not print message in DCC output command window
    """

    pass


@reroute
@abstract
def input_comment(title, label, text=''):
    """
    Shows a comment input dialog that users can use to input text.

    :param str title: text that is displayed in the title bar of the dialog
    :param str label: text which is shown to the user telling them what kind of text they need to input
    :param str text: default text which is placed in the comment section
    :return: Tuple containing as the first element the text typed by the user and as the second argument a boolean
        that will be True if the user clicked on the Ok button or False if the user cancelled the operation
    :rtype: tuple(str, bool)
    """

    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains ArtellaDrive Qt related utils classes and functions
"""

from __future__ import print_function, division, absolute_import

import os

from artella import dcc
from artella.core import utils, qtutils

if qtutils.QT_AVAILABLE:
    from artella.externals.Qt import QtGui


class ResourceTypes(object):
    ICON = 'icon'
    PIXMAP = 'pixmap'
    STYLE = 'style'


_RESOURCES_PATHS = list()
_RESOURCES_CACHE = {
    ResourceTypes.ICON: dict(), ResourceTypes.PIXMAP: dict(), ResourceTypes.STYLE: dict()
}


def register_resources_path(resources_path):
    """
    Registers a path where resources will be searched
    :param str resources_path: Path to search resources in
    """

    if not resources_path or not os.path.isdir(resources_path):
        return

    resources_path = utils.clean_path(resources_path)
    if resources_path in _RESOURCES_PATHS:
        return

    _RESOURCES_PATHS.append(resources_path)

    dcc.register_dcc_resource_path(resources_path)
    icons_path = os.path.join(resources_path, 'icons')
    if os.path.isdir(icons_path):
        dcc.register_dcc_resource_path(icons_path)


def get(resource_type, name, extension, **kwargs):
    """
    Internal function that returns a resource based on its type
    :param resource_type:
    :return: object
    """

    color = kwargs.get('color', '')
    file_name = '{}{}'.format(name, extension)
    file_key = '{}{}'.format(file_name.lower(), color)

    if not qtutils.QT_AVAILABLE or not _RESOURCES_PATHS or resource_type not in _RESOURCES_CACHE:
        return None

    if not extension.startswith('.'):
        extension = '.{}'.format(extension)

    if file_key in _RESOURCES_CACHE[resource_type]:
        return _RESOURCES_CACHE[resource_type][file_key]

    for resource_path in _RESOURCES_PATHS:
        for root, dirs, files in os.walk(resource_path):
            for path in files:
                path_name, path_ext = os.path.splitext(path)
                if path_name == name and path_ext == extension:
                    res_path = os.path.join(root, path)
                    new_res = None
                    if resource_type == ResourceTypes.ICON:
                        new_res = qtutils.icon(res_path, color=color)
                    elif resource_type == ResourceTypes.PIXMAP:
                        new_res = qtutils.pixmap(res_path, color=color)
                    elif resource_type == ResourceTypes.STYLE:
                        new_res = qtutils.style(res_path)
                    if not new_res:
                        return None

                    if hasattr(new_res, 'isNull'):
                        if not new_res.isNull():
                            _RESOURCES_CACHE[resource_type][file_key] = new_res
                    else:
                        _RESOURCES_CACHE[resource_type][file_key] = new_res

                    return new_res

    return None


def icon(name, extension='png', color=None):
    """
    Returns Artella icon
    :param name:
    :param extension:
    :param color:
    :return: QIcon
    """

    new_icon = get(ResourceTypes.ICON, name=name, extension=extension, color=color)
    if not new_icon:
        return None if not qtutils.QT_AVAILABLE else QtGui.QIcon()

    return new_icon


def pixmap(name, extension='png', color=None):
    """
    Returns Artella pixmap resource
    :param name:
    :param extension:
    :param color:
    :return:
    """

    new_pixmap = get(ResourceTypes.PIXMAP, name=name, extension=extension, color=color)
    if not new_pixmap:
        return None if not qtutils.QT_AVAILABLE else QtGui.QPixmap()

    return new_pixmap


def style(name, extension='css'):
    """
    Returns Artella style resource
    :param name:
    :param extension:
    :return:
    """

    new_style = get(ResourceTypes.STYLE, name=name, extension=extension)
    if not new_style:
        return None

    return new_style



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains DCC abstract scene file parser implementation
"""

from __future__ import print_function, division, absolute_import

from artella import dcc
from artella.core.dcc import reroute
from artella.core.utils import abstract, add_metaclass


class _MetaDccParser(type):

    def __call__(cls, *args, **kwargs):
        if dcc.is_maya():
            from artella.dccs.maya import parser
            return type.__call__(parser.MayaSceneParser, *args, **kwargs)
        else:
            return type.__call__(BaseSceneParser, *args, **kwargs)


class AbstractSceneParser(object):
    """
    Class that defines basic scene parser abstract functions
    """

    @abstract
    @reroute
    def parse(self, file_paths=None):
        """
        Parses all the contents of the given file path looking for file paths

        :param file_paths: list(str) or str or None: Absolute local file path of the DCC file we want to parse.
            If not given, current opened DCC scene file path will be used
        :return:
        """

        pass

    @abstract
    @reroute
    def update_paths(self, file_paths=None):
        """
        Converts all file path of the given DCC file to make sure they point to valid Artella file paths

        :param file_paths: list(str) or str or None: Absolute local file path of the DCC file we want to parse.
            If not given, current opened DCC scene file path will be used
        :return:
        """

        pass


class BaseSceneParser(AbstractSceneParser):
    def parse(self, file_paths=None):
        return list()

    def update_paths(self, file_paths=None):
        return False


@add_metaclass(_MetaDccParser)
class Parser(AbstractSceneParser):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains DCC abstract window implementation
"""

from __future__ import print_function, division, absolute_import

from artella import dcc
from artella.core import utils, qtutils, resource
from artella.widgets import theme

if qtutils.QT_AVAILABLE:
    from artella.externals.Qt import QtCore, QtWidgets


class _MetaWindow(type):

    def __call__(cls, *args, **kwargs):
        if dcc.is_maya():
            from artella.dccs.maya import window
            return window.MayaWindow
        else:
            return BaseWindow


class AbstractWindow(object):

    @utils.abstract
    def get_main_layout(self):
        pass

    @utils.abstract
    def setup_ui(self):
        pass


if not qtutils.QT_AVAILABLE:
    class BaseWindow(object):
        pass
else:
    class BaseWindow(QtWidgets.QMainWindow):
        def __init__(self, parent=None, **kwargs):
            if not parent:
                from artella import dcc
                parent = dcc.get_main_window()

            super(BaseWindow, self).__init__(parent, **kwargs)

            self._pos_anim = QtCore.QPropertyAnimation(self)
            self._pos_anim.setTargetObject(self)
            self._pos_anim.setEasingCurve(QtCore.QEasingCurve.OutCubic)
            self._pos_anim.setDuration(300)
            self._pos_anim.setPropertyName(b'pos')

            self._opacity_anim = QtCore.QPropertyAnimation()
            self._opacity_anim.setTargetObject(self)
            self._opacity_anim.setEasingCurve(QtCore.QEasingCurve.OutCubic)
            self._opacity_anim.setDuration(300)
            self._opacity_anim.setPropertyName(b'windowOpacity')
            self._opacity_anim.setStartValue(0.0)
            self._opacity_anim.setEndValue(1.0)

            self.setup_ui()
            theme.theme().apply(self)

            self._fade_in()

        def get_main_layout(self):
            main_layout = QtWidgets.QVBoxLayout()
            main_layout.setContentsMargins(0, 0, 0, 0)
            main_layout.setSpacing(0)

            return main_layout

        def setup_ui(self):
            main_widget = QtWidgets.QWidget()
            self.main_layout = self.get_main_layout()
            main_widget.setLayout(self.main_layout)
            self.setCentralWidget(main_widget)

            artella_frame = QtWidgets.QFrame()
            artella_frame.setObjectName('artellaFrame')
            artella_frame_layout = QtWidgets.QHBoxLayout()
            artella_frame.setLayout(artella_frame_layout)

            artella_header = QtWidgets.QLabel()
            artella_header_pixmap = resource.pixmap('artella_header')
            artella_header.setPixmap(artella_header_pixmap)
            artella_frame_layout.addStretch()
            artella_frame_layout.addWidget(artella_header)
            artella_frame_layout.addStretch()

            self.main_layout.addWidget(artella_frame)

        def fade_close(self):
            self._fade_out()

        # =================================================================================================================
        # INTERNAL
        # =================================================================================================================

        def _fade_out(self):
            self._opacity_anim.finished.connect(self.close)
            self._pos_anim.setDirection(QtCore.QAbstractAnimation.Backward)
            self._pos_anim.start()
            self._opacity_anim.setDirection(QtCore.QAbstractAnimation.Backward)
            self._opacity_anim.start()

        def _fade_in(self):
            self._pos_anim.start()
            self._opacity_anim.start()


@utils.add_metaclass(_MetaWindow)
class Window(AbstractWindow):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains DCC abstract progress bar implementation
"""

from __future__ import print_function, division, absolute_import

from artella import dcc
from artella.core.utils import abstract, add_metaclass


class _MetaProgressBar(type):

    def __call__(cls, *args, **kwargs):
        if dcc.is_maya():
            from artella.dccs.maya import progress as maya_progress
            return type.__call__(maya_progress.MayaProgressBar, *args, **kwargs)
        else:
            return type.__call__(AbstractProgressBar, *args, **kwargs)


class AbstractProgressBar(object):
    """
    Class that defines basic progress bar abstract functions
    """

    @abstract
    def can_be_interrupted(self):
        """
        Returns whether or not DCC progress bar can be interrupted or not
        :return: True if the progress bar can be interrupted by the user; False otherwise.
        :rtype: bool
        """

        pass

    @abstract
    def is_cancelled(self):
        """
        Returns whether or not DCC progress bar has been cancelled by the user

        :return: True if the progres bar has been cancelled by the user; False otherwise.
        :rtype: bool
        """

        pass

    @abstract
    def get_max_progress_value(self):
        """
        Returns the maximum value of the progress bar

        :return: Maximum value progress bar can accept
        :rtype: int
        """

        pass

    def get_min_progress_value(self):
        """
        Returns the minimum value of the progress bar

        :return: Minimum value progress bar can accept
        :rtype: int
        """

        pass

    @abstract
    def set_min_progress_value(self, min_value):
        """
        Sets the minimum value of the progress bar

        :param int min_value: Minimum value progress bar can accept
        """

        pass

    def set_max_progress_value(self, max_value):
        """
        Sets the maximum value of the progress bar

        :param int max_value: Maximum value progress bar can accept
        """

        pass

    @abstract
    def get_progress_value(self):
        """
        Returns current progress value of the progress bar

        :return: current progress value
        :rtype: int
        """

        pass

    @abstract
    def set_progress_value(self, value):
        """
        Sets the current progress value of the progress bar

        :param int value: current progress value
        """

        pass

    @abstract
    def increment_value(self, increment=1):
        """
        Increments current progress value with the given increment

        :param int increment: Increment step we want to apply to current progress bar value
        """

        pass

    @abstract
    def get_status(self):
        """
        Returns current status text of the progress bar

        :return: status text of the progress bar
        :rtype: str
        """

        pass

    @abstract
    def set_status(self, status_text):
        """
        Sets current status text of the progress bar

        :param str status_text: text used by progress bar
        """

        pass

    @abstract
    def start(self, title='', status='', min_count=0, max_count=100):
        """
        Starts progress bar execution
        :param str title: Title of the progress bar (optional)
        :param str status: Initial text used by progress bar (optional)
        :param int min_count: Initial minimum progress bar value (optional)
        :param int max_count: Initial maximum progress bar value (optional)
        """

        pass

    @abstract
    def end(self):
        """
        Ends progress bar execution
        """

        pass


class BaseProgressBar(object):
    def can_be_interrupted(self):
        return False

    def is_cancelled(self):
        return True

    @abstract
    def get_max_progress_value(self):
        return 100

    def get_min_progress_value(self):
        return 0

    @abstract
    def set_min_progress_value(self, min_value):
        pass

    def set_max_progress_value(self, max_value):
        pass

    def get_progress_value(self):
        pass

    def set_progress_value(self, value):
        pass

    def increment_value(self, increment=1):
        pass

    def get_status(self):
        pass

    def set_status(self, status_text):
        pass

    @abstract
    def start(self, title='', status='', min_count=0, max_count=100):
        pass

    @abstract
    def end(self):
        pass


@add_metaclass(_MetaProgressBar)
class ProgressBar(AbstractProgressBar):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains Artella Plugins framework implementation
"""

from __future__ import print_function, division, absolute_import

import logging
import inspect

from artella import dcc
from artella.core import dccplugin

logger = logging.getLogger('artella')


class ArtellaPlugin(object):

    ID = ''                 # Unique ID of the Artella Plugin
    INDEX = -1              # Index of the plugin. This index will control the instantiation order
    VERSION = None          # Version of the Artella Plugin
    PACKAGE = None          # Package Artella Plugin belongs to

    def __init__(self, config_dict=None, manager=None):

        self._config_dict = config_dict or dict()
        self._manager = manager
        self._stats = ArtellaPluginStats(self)
        self._loaded = False

        self.init()

    def is_loaded(self):
        """
        Returns whether or not this plugin is loaded
        :return: True if the plugin is loaded; False otherwise.
        :rtype: bool
        """

        return self._loaded

    def init(self):
        """
        Function that is called when plugin is instantiated
        """

        if self._loaded:
            dcc.execute_deferred(self.cleanup)

        dcc.execute_deferred(self.init_ui)

        self._loaded = True

    def init_ui(self):
        """
        Function that initializes plugin UI related functionality
        """

        main_menu = dccplugin.DccPlugin().get_main_menu()
        if not main_menu:
            return

        plugin_menu = self._config_dict.get('menu')
        plugin_package = self._config_dict.get('package', 'Artella')
        plugin_dccs = self._config_dict.get('dcc', list())

        can_load_plugin = True
        if plugin_dccs:
            can_load_plugin = dcc.name() in plugin_dccs

        if can_load_plugin:
            if plugin_menu and 'label' in plugin_menu:
                menu_parents = plugin_menu.get('parents', None)
                if menu_parents:
                    current_parent = 'Artella' if dcc.check_menu_exists('Artella') else None
                    for menu_parent in menu_parents:
                        if not dcc.check_menu_exists(menu_parent):

                            # TODO: Before More Artella menu addition we force the creation of a
                            # TODO: separator. We should find a way to avoid hardcoded this.
                            icon = ''
                            if menu_parent == 'More Artella':
                                dcc.add_menu_separator('Artella')
                                icon = 'artella.png'

                            dcc.add_sub_menu_item(menu_parent, parent_menu=current_parent, icon=icon)
                        current_parent = menu_parent

                menu_label = plugin_menu['label']
                menu_command = plugin_menu['command']
                menu_icon = self._config_dict.get('icon', '')
                if menu_parents:
                    menu_parent = menu_parents[-1]
                    dcc.add_menu_item(menu_label, menu_command, menu_parent, icon=menu_icon)
                else:
                    dcc.add_menu_item(menu_label, menu_command, main_menu, icon=menu_icon)

    def cleanup(self):
        """
        Function that is called when the plugin is disabled
        """

        plugin_menu = self._config_dict.get('menu')
        menu = dcc.get_menu('Artella')

        if plugin_menu and menu:
            menu_label = plugin_menu['label']
            dcc.remove_menu_item(menu_label, menu)

        self._loaded = False

    @property
    def manager(self):
        """
        Returns Artella Plugins manager instance that is owner of this plugin

        :return: Artella Plugin manager instance that owns this plugin
        :rtype: ArtellaPluginsManager
        """

        return self._manager

    @property
    def stats(self):
        """
        Returns Artella Plugin Stats instance that stores useful data related with the plugin

        :return: Artella Plugin Stats instance with useful data of the plugin
        :rtype: ArtellaPluginStats
        """

        return self._stats


class ArtellaPluginStats(object):
    def __init__(self, plugin):
        self._plugin = plugin
        self._id = self._plugin.ID
        self._start_time = 0.0
        self._end_time = 0.0
        self._execution_time = 0.0
        self._info = dict()
        self._init()

    @property
    def start_time(self):
        """
        Returns the time the plugin was executed

        :return: Time the plugin was executed
        :rtype: str
        """

        return self._start_time

    @start_time.setter
    def start_time(self, value):
        """
        Sets the time the plugin was executed

        :param str value: Time the plugin was executed
        """

        self._start_time = value

    @property
    def end_time(self):
        """
        Returns the time the plugin ended its execution

        :return: Time the plugin was executed
        :rtype: str
        """

        return self._end_time

    @end_time.setter
    def end_time(self, value):
        """
        Sets the time the plugin ended its execution

        :param str value: Time the plugin ended its execution
        """

        self._end_time = value

    @property
    def execution_time(self):
        """
        Returns the the total amount of execution time of the plugin

        :return: Total execution time of the plugin during current session in seconds
        :rtype: float
        """

        return self._execution_time

    def _init(self):
        """
        Internal function that initializes info for the plugin and its environment
        """

        from artella import dcc

        self._info.update({
            'name': self._plugin.__class__.__name__,
            'module': self._plugin.__class__.__module__,
            'filepath': inspect.getfile(self._plugin.__class__),
            'id': self._id,
            'application': dcc.name()
        })
